"""
Heterogeneous atmospheres.
"""

from __future__ import annotations

import logging
from collections import abc as cabc

import attrs
import mitsuba as mi
import numpy as np
import pint

from ._core import AbstractHeterogeneousAtmosphere, atmosphere_factory
from ._molecular import MolecularAtmosphere
from ._particle_layer import ParticleLayer
from ..core import traverse
from ..phase import BlendPhaseFunction, PhaseFunction
from ...attrs import define, documented
from ...contexts import KernelContext
from ...kernel import SearchSceneParameter
from ...radprops import AbsorptionDatabase, ZGrid
from ...spectral.index import SpectralIndex
from ...units import unit_context_config as ucc
from ...units import unit_registry as ureg
from ...util.misc import cache_by_id

logger = logging.getLogger(__name__)


def _molecular_converter(value):
    if isinstance(value, cabc.MutableMapping) and ("type" not in value):
        value["type"] = "molecular"
    return atmosphere_factory.convert(value, allowed_cls=MolecularAtmosphere)


def _particle_layer_converter(value):
    if not value:
        return []

    if not isinstance(value, (list, tuple)):
        return _particle_layer_converter([value])

    else:
        result = []

        for element in value:
            if isinstance(element, cabc.MutableMapping) and ("type" not in element):
                element["type"] = "particle_layer"
            result.append(
                atmosphere_factory.convert(element, allowed_cls=ParticleLayer)
            )

        return result


@define(eq=False, slots=False)
class HeterogeneousAtmosphere(AbstractHeterogeneousAtmosphere):
    """
    Heterogeneous atmosphere scene element [``heterogeneous``].
    """

    molecular_atmosphere: MolecularAtmosphere | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(_molecular_converter),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(MolecularAtmosphere)
            ),
        ),
        doc="Molecular atmosphere. May be specified as a dictionary interpreted "
        'by :data:`.atmosphere_factory`; in that case, the ``"type"`` parameter '
        'may be omitted and will automatically be set to ``"molecular"``.',
        type=".MolecularAtmosphere or None",
        init_type=".MolecularAtmosphere or dict, optional",
        default="None",
    )

    @molecular_atmosphere.validator
    def _molecular_atmosphere_validator(self, attribute, value):
        if value is None:
            return

        if value.scale is not None:
            raise ValueError(
                f"while validating {attribute.name}: components cannot be "
                "scaled individually"
            )

    particle_layers: list[ParticleLayer] = documented(
        attrs.field(
            factory=list,
            converter=_particle_layer_converter,
            validator=attrs.validators.deep_iterable(
                attrs.validators.instance_of(ParticleLayer)
            ),
        ),
        doc="List of particle layers. Elements may be specified as "
        "dictionaries interpreted by :data:`.atmosphere_factory`; in that "
        "case, the ``type`` parameter may be omitted and will automatically "
        'be set to ``"particle_layer"``.',
        type="list of .ParticleLayer",
        init_type="list of .ParticleLayer, optional",
        default="[]",
    )

    @particle_layers.validator
    def _particle_layers_validator(self, attribute, value):
        if not all(component.scale is None for component in value):
            raise ValueError(
                f"while validating {attribute.name}: components cannot be "
                "scaled individually"
            )

    _zgrid: ZGrid = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(
                lambda x: ZGrid(x) if not isinstance(x, ZGrid) else x
            ),
            validator=attrs.validators.optional(attrs.validators.instance_of(ZGrid)),
            repr=False,
        ),
        doc="A high-resolution layer altitude mesh on which the radiative "
        "properties of the components are interpolated. If unset, a default grid "
        "with one layer per 100 m (or 10 layers if the atmosphere object "
        "height is less than 100 m) is used.",
        type=".ZGrid",
        init_type=".ZGrid, optional",
    )

    @property
    def components(self) -> list[MolecularAtmosphere | ParticleLayer]:
        """
        Returns
        -------
        list of .AbstractHeterogeneousAtmosphere
            The list of all registered atmospheric components.
        """
        result = [self.molecular_atmosphere] if self.molecular_atmosphere else []
        result.extend(self.particle_layers)
        return result

    def update(self):
        # Inherit docstring
        if not self.components:
            raise ValueError("HeterogeneousAtmosphere must have at least one component")

        # Force component IDs
        for i, component in enumerate(self.components):
            component.update()
            component.id = f"{self.id}_component_{i}"
            logger.debug(
                "Override the geometry of component '%s' to %s (was %s)",
                component.id,
                self.geometry,
                component.geometry,
            )
            component.geometry = self.geometry

    # --------------------------------------------------------------------------
    #              Spatial extension and thermophysical properties
    # --------------------------------------------------------------------------

    @property
    def zgrid(self) -> ZGrid:
        # Inherit docstring
        return self._zgrid

    @property
    def bottom(self) -> pint.Quantity:
        # Inherit docstring
        return min([component.bottom for component in self.components])

    @property
    def top(self) -> pint.Quantity:
        # Inherit docstring
        return max([component.top for component in self.components])

    def eval_mfp(self, ctx: KernelContext) -> pint.Quantity:
        # Inherit docstring
        mfp = [component.eval_mfp(ctx=ctx) for component in self.components]
        return max(mfp)

    # --------------------------------------------------------------------------
    #                       Radiative properties
    # --------------------------------------------------------------------------

    @property
    def absorption_data(self) -> AbsorptionDatabase | None:
        # Inherit docstring
        if self.molecular_atmosphere is not None:
            return self.molecular_atmosphere._absorption_data
        else:
            return None

    def eval_albedo(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        if zgrid is not None and zgrid is not self.geometry.zgrid:
            raise ValueError("zgrid must be left unset or set to self.geometry.zgrid")

        units = ucc.get("collision_coefficient")
        sigma_s = self.eval_sigma_s(si).m_as(units)
        sigma_t = self.eval_sigma_t(si).m_as(units)
        albedo = np.zeros_like(sigma_s)
        np.divide(sigma_s, sigma_t, where=sigma_t != 0.0, out=albedo)

        return albedo * ureg.dimensionless

    @cache_by_id
    def _eval_sigma_t_impl(self, si: SpectralIndex) -> pint.Quantity:
        result = np.zeros((len(self.components), self.geometry.zgrid.n_layers))
        sigma_units = ucc.get("collision_coefficient")

        # Evaluate extinction for current component
        for i, component in enumerate(self.components):
            result[i] = component.eval_sigma_t(si, self.geometry.zgrid).m_as(
                sigma_units
            )

        return result * sigma_units

    def eval_sigma_t(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        if zgrid is not None and zgrid is not self.geometry.zgrid:
            raise ValueError("zgrid must be left unset or set to self.geometry.zgrid")
        return self._eval_sigma_t_impl(si).sum(axis=0)

    def eval_sigma_a(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        if zgrid is not None and zgrid is not self.geometry.zgrid:
            raise ValueError("zgrid must be left unset or set to self.geometry.zgrid")
        return self.eval_sigma_t(si) - self.eval_sigma_s(si)

    @cache_by_id
    def _eval_sigma_s_impl(self, si: SpectralIndex) -> pint.Quantity:
        result = np.zeros((len(self.components), self.geometry.zgrid.n_layers))
        sigma_units = ucc.get("collision_coefficient")

        # Evaluate scattering coefficient for current component
        for i, component in enumerate(self.components):
            result[i] = component.eval_sigma_s(si, self.geometry.zgrid).m_as(
                sigma_units
            )

        return result * sigma_units

    def _eval_sigma_s_component(
        self, si: SpectralIndex, n_component: int
    ) -> pint.Quantity:
        return self._eval_sigma_s_impl(si)[n_component]

    def eval_sigma_s(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        if zgrid is not None and zgrid is not self.geometry.zgrid:
            raise ValueError("zgrid must be left unset or set to self.geometry.zgrid")
        return self._eval_sigma_s_impl(si).sum(axis=0)

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def phase(self) -> PhaseFunction:
        # Inherit docstring
        if len(self.components) == 1:
            return self.components[0].phase

        else:
            components, weights = [], []
            sigma_units = ucc.get("collision_coefficient")

            for i, component in enumerate(self.components):
                components.append(component.phase)

                def eval_sigma_s(si: SpectralIndex, n_component: int = i) -> np.ndarray:
                    return self._eval_sigma_s_component(si, n_component).m_as(
                        sigma_units
                    )

                weights.append(eval_sigma_s)

            return BlendPhaseFunction(
                components=components, weights=weights, geometry=self.geometry
            )

    @property
    def _template_phase(self) -> dict:
        # Inherit docstring
        return traverse(self.phase)[0].data

    @property
    def _params_phase(self) -> dict:
        # Inherit docstring

        umap = traverse(self.phase)[1].data

        # Add prefix and lookup strategy to all entries
        result = {}

        for uparam_key, uparam in umap.items():
            result[uparam_key] = attrs.evolve(
                uparam,
                search=SearchSceneParameter(
                    node_type=mi.Medium,
                    node_id=self.medium_id,
                    parameter_relpath=f"phase_function.{uparam_key}",
                ),
            )

        return result



from __future__ import annotations

from pathlib import Path

import attrs
import xarray as xr

from ._asset_manager import asset_manager
from ..attrs import define
from ..config import SOURCE_DIR, settings
from ..typing import PathLike


def _validator_dir_exists(instance, attribute, value):
    if not value.is_dir():
        raise NotADirectoryError(value)


@define
class FileResolver:
    """
    This class resolves paths relative to a list of locations on disk.
    Locations are looked up in order upon calling the :meth:`.resolve` method.
    If a lookup is successful, the resolved absolute path is returned;
    otherwise, the input path is returned unchanged.

    Parameters
    ----------
    paths : list of path-likes, optional
        A list of search directories. Each entry must exist.

    Notes
    -----
    * By default, the unique file resolver instance
      :data:`eradiate.fresolver <.fresolver>` is initialized with the
      installation path of the asset manager (see :class:`.AssetManager`) and,
      when in dev mode (see :data:`SOURCE_DIR <eradiate.config.SOURCE_DIR>`),
      the path to the location of test files.
    * The current working directory is not appended to the file resolver by
      default. It can however optionally be added temporarily when calling
      :meth:`resolve` (and wrappers such as :meth:`load_dataset`). This is
      useful when using the ``strict`` mode, which raises if the requested path
      cannot be resolved to an existing file.


    Examples
    --------
    A single instance of the file resolver is available as
    :data:`eradiate.fresolver <.fresolver>`:

    >>> from eradiate import fresolver

    To add a path to the file resolver, use the :meth:`~.FileResolver.append`
    or :meth:`~.FileResolver.prepend` methods, *e.g.*:

    >>> fresolver.append("some/path/on/the/drive")

    To resolve a relative path, use the :meth:`~.FileResolver.resolve` method:

    >>> fresolver.resolve("srf/sentinel_2a-msi-3.nc")

    If the path is expected to point to an existing dataset file, the
    :meth:`~.FileResolver.load_dataset` method can be used to load it immediately:

    >>> fresolver.load_dataset("srf/sentinel_2a-msi-3.nc")
    """

    paths: list[Path] = attrs.field(
        factory=list,
        converter=lambda value: [Path(x).resolve() for x in value],
        validator=attrs.validators.deep_iterable(_validator_dir_exists),
    )

    def append(self, path: PathLike, avoid_duplicates: bool = True) -> None:
        """
        Append an entry to the end of the list of search paths.

        Parameters
        ----------
        path : path-like
            Path to append. The location must exist on disk.

        avoid_duplicates : bool, optional
            If ``True``, do not append again a path that is already registered.

        See Also
        --------
        :meth:`.prepend`
        """
        path = Path(path).resolve()
        if not path.is_dir():
            raise NotADirectoryError(f"{path}")

        if avoid_duplicates and path in self.paths:
            return

        self.paths.append(path)

    def clear(self) -> None:
        """
        Clear the list of search paths.
        """
        self.paths.clear()

    def prepend(self, path: PathLike, avoid_duplicates: bool = True) -> None:
        """
        Prepend an entry at the beginning of the list of search paths.

        Parameters
        ----------
        path : path-like
            Path to prepend. The location must exist on disk.

        avoid_duplicates : bool, optional
            If ``True``, do not prepend again a path that is already registered.

        See Also
        --------
        :meth:`.append`
        """
        path = Path(path).resolve()
        if not path.is_dir():
            raise NotADirectoryError(f"{path}")

        if avoid_duplicates and path in self.paths:
            return

        self.paths.insert(0, path)

    def resolve(self, path: PathLike, strict: bool = False, cwd: bool = False) -> Path:
        """
        Resolve a path: search all registered locations in order. If no file is
        found in any of the registered location, ``path`` is returned unchanged.

        Parameters
        ----------
        path : path-like
            Path to be resolved.

        strict : bool, default: False
            If ``True``, resolution failure will raise.

        cwd : bool, default: False
            If ``True``, check first if a file relative to the current working
            directory exists.

        Returns
        -------
        Path
            Resolved path

        Raises
        ------
        FileNotFoundError
            If ``strict`` is ``True`` and ``path`` was not found in one of the
            registered directories.
        """
        path = Path(path)

        if not path.is_absolute():
            for base in self.paths if not cwd else [Path.cwd()] + self.paths:
                combined = base / path
                if combined.exists():
                    return combined

        if strict:
            raise FileNotFoundError(path)

        return path

    def load_dataset(
        self, path: PathLike, strict: bool = False, cwd: bool = False
    ) -> xr.Dataset:
        """
        Chain :meth:`resolve` and :func:`xarray.load_dataset`.

        Parameters
        ----------
        path : path-like
            Path to be resolved.

        strict : bool, default: False
            If ``True``, resolution failure will raise.

        cwd : bool, default: False
            If ``True``, check first if a file relative to the current working
            directory exists.

        Returns
        -------
        Dataset
        """
        fname = self.resolve(path, strict=strict, cwd=cwd)
        return xr.load_dataset(fname)

    def info(self, show: bool = False) -> dict | None:
        """
        Collect information about the file resolver.

        Parameters
        ----------
        show : bool
            If ``True``, display information to the terminal. Otherwise, return
            it as a dictionary.

        Returns
        -------
        dict or None
        """
        if show:
            title = "File resolver"
            print(title)
            print("-" * len(title))

            for path in self.paths:
                print(f"• {path}")

            return None

        else:
            return {"paths": self.paths}


#: Unique file resolver instance (exposed as :data:`eradiate.fresolver`)
fresolver = FileResolver(settings["path"])
fresolver.append(asset_manager.install_dir)
if SOURCE_DIR:
    fresolver.append(SOURCE_DIR / "resources/data")



from __future__ import annotations

import attrs
import mitsuba as mi

from ._core import BSDF
from ..core import traverse
from ..spectra import Spectrum, spectrum_factory
from ... import validators
from ...attrs import define, documented
from ...kernel import SceneParameter, SearchSceneParameter


@define(eq=False, slots=False)
class CheckerboardBSDF(BSDF):
    """
    Checkerboard BSDF [``checkerboard``].

    This class defines a Lambertian BSDF textured with a checkerboard pattern.
    """

    reflectance_a: Spectrum = documented(
        attrs.field(
            default=0.2,
            converter=spectrum_factory.converter("reflectance"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("reflectance"),
            ],
        ),
        doc="Reflectance spectrum. Can be initialized with a dictionary "
        "processed by :data:`.spectrum_factory`.",
        type=".Spectrum",
        init_type=".Spectrum or dict or float",
        default="0.2",
    )

    reflectance_b: Spectrum = documented(
        attrs.field(
            default=0.8,
            converter=spectrum_factory.converter("reflectance"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("reflectance"),
            ],
        ),
        doc="Reflectance spectrum. Can be initialized with a dictionary "
        "processed by :data:`.spectrum_factory`.",
        type=".Spectrum",
        init_type=":class:`.Spectrum` or dict or float",
        default="0.8",
    )

    scale_pattern: float = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(float),
            validator=attrs.validators.optional(attrs.validators.instance_of(float)),
        ),
        doc="Scaling factor for the checkerboard pattern. The higher the value, "
        "the more checkerboard patterns will fit on the surface to which this "
        "reflection model is attached.",
        type="float or None",
        init_type="float, optional",
    )

    @property
    def template(self) -> dict:
        # Inherit docstring

        result = {"type": "diffuse", "reflectance.type": "checkerboard"}

        if self.id is not None:
            result["id"] = self.id

        if self.scale_pattern is not None:
            result["reflectance.to_uv"] = mi.ScalarTransform4f.scale(self.scale_pattern)

        for obj_key, obj_values in {
            "color0": traverse(self.reflectance_a)[0],
            "color1": traverse(self.reflectance_b)[0],
        }.items():
            for key, value in obj_values.items():
                result[f"reflectance.{obj_key}.{key}"] = value

        return result

    @property
    def params(self) -> dict[str, SceneParameter]:
        # Inherit docstring

        result = {}

        for obj_key, obj_params in {
            "color0": traverse(self.reflectance_a)[1].data,
            "color1": traverse(self.reflectance_b)[1].data,
        }.items():
            for key, param in obj_params.items():
                result[f"reflectance.{obj_key}.{key}"] = attrs.evolve(
                    param,
                    search=SearchSceneParameter(
                        node_type=mi.BSDF,
                        node_id=self.id,
                        parameter_relpath=f"reflectance.{obj_key}.{key}",
                    )
                    if self.id is not None
                    else None,
                )

        return result



"""Frame and angle manipulation utilities."""

from __future__ import annotations

import enum
import typing as t

import aenum
import numpy as np
import pint

from .units import unit_registry as ureg


class AzimuthConvention(enum.Enum):
    """
    An enumeration of azimuth convention names associated with the corresponding
    (origin offset, orientation) pair with respect to the *East right*
    convention. The origin offset is expressed in radian and the orientation is
    the angle value multiplier (±1).
    """

    EAST_RIGHT = (0.0, 1)  #: East right
    EAST_LEFT = (0.0, -1)  #: East left
    NORTH_RIGHT = (0.5 * np.pi, 1)  #: North right
    NORTH_LEFT = (0.5 * np.pi, -1)  #: North left
    WEST_RIGHT = (np.pi, 1)  #: West right
    WEST_LEFT = (np.pi, -1)  #: West left
    SOUTH_RIGHT = (1.5 * np.pi, 1)  #: South right
    SOUTH_LEFT = (1.5 * np.pi, -1)  #: South left

    @staticmethod
    def convert(value: t.Any) -> AzimuthConvention:
        """
        Attempt conversion of a value to an :class:`.AzimuthConvention`
        instance. The conversion protocol is as follows:

        * If ``value`` is a string, it is converted to upper case and passed to
          the indexing operator of :class:`.AzimuthConvention`.
        * If ``value`` is an :class:`.AzimuthConvention` instance, it is returned
          without change.
        * Otherwise, the method raises an exception.

        Parameters
        ----------
        value
            Value to attempt conversion of.

        Returns
        -------
        Converted value

        Raises
        ------
        TypeError
            If no conversion protocol exists for ``value``.
        """
        if isinstance(value, str):
            return AzimuthConvention[value.upper()]
        elif isinstance(value, AzimuthConvention):
            return value
        else:
            raise TypeError(
                f"Cannot convert a {type(value)} instance to AzimuthConvention"
            )

    @classmethod
    def register(cls, name: str, value: tuple[float, float]) -> None:
        """
        Register a new angular convention.

        Parameters
        ----------
        name : str
            Name of the registered convention. Should be uppercase, without
            whitespace.

        value : tuple
            An (offset, orientation) 2-tuple. The offset is the angular offset
            of the registered convention in radian. The offset is an integer
            equal to 1 (right-hand/counter-clockwise convention) or -1
            (left-hand/clockwise convention).
        """
        aenum.extend_enum(cls, name.upper(), value)


def normalize_azimuth(angles: np.typing.ArrayLike, inplace: bool = False) -> np.ndarray:
    """
    Normalize azimuth values to the [0, 2π[ interval.

    Parameters
    ----------
    angles : array-like
        A sequence of azimuth values [rad].

    inplace : bool, optional, default: False
        If ``True``, perform the conversion in-place. This will mutate the
        `angles` array; otherwise, this function operates on a copy.

    Returns
    -------
    ndarray
        Azimuth angle values normalized to the [0, 2π[ interval [rad].

    Warnings
    --------
    This function does *not* apply unit conversion automatically: angle values
    must be supplied in radians and are returned as plain Numpy arrays.
    """
    result = angles if inplace else np.copy(angles)
    result %= 2.0 * np.pi

    # Snap close-to-2π values to 0 to compensate for numerical precision-related
    # unwanted shifts (may happen with angles computed from directions)
    snapped = np.where(
        np.isclose(result, 2.0 * np.pi, rtol=0.0, atol=1e-6 * np.pi), 0.0, result
    )

    if inplace:
        result[:] = snapped
    else:
        result = snapped

    return result


def transform_azimuth(
    angles: np.typing.ArrayLike,
    from_convention: AzimuthConvention | str = AzimuthConvention.EAST_RIGHT,
    to_convention: AzimuthConvention | str = AzimuthConvention.EAST_RIGHT,
    normalize: bool = True,
    inplace: bool = False,
) -> np.ndarray:
    """
    Convert azimuth values expressed with a given convention to another. The
    East right convention is used as a pivot.

    Parameters
    ----------
    angles : array-like
        A sequence of azimuth values [rad].

    from_convention : .AzimuthConvention or str, optional, default: .AzimuthConvention.EAST_RIGHT
        Source azimuth angle convention. If a string is passed, it will be
        converted to a :class:`.AzimuthConvention`.

    to_convention : .AzimuthConvention or str, optional, default: .AzimuthConvention.EAST_RIGHT
        Target azimuth angle convention. If a string is passed, it will be
        converted to a :class:`.AzimuthConvention`.

    normalize : bool, optional, default: True
        If ``True``, normalize returned angle values within the [0, 2π[
        interval.

    inplace : bool, optional, default: False
        If ``True``, perform the conversion in-place. This will mutate the
        `angles` array; otherwise, this function operates on a copy.

    Returns
    -------
    ndarray
        Azimuth angle values converted to the East right convention [rad].

    Warnings
    --------
    This function does *not* apply unit conversion automatically: angle values
    must be supplied in radians and are returned as plain Numpy arrays.
    """
    result = angles if inplace else np.copy(angles)

    from_convention = AzimuthConvention.convert(from_convention)
    to_convention = AzimuthConvention.convert(to_convention)

    if from_convention is not to_convention:
        from_offset, from_orientation = from_convention.value
        to_offset, to_orientation = to_convention.value

        # Transform to East right
        result *= from_orientation
        result += from_offset

        # Transform to target convention
        result -= to_offset
        result *= to_orientation

    if normalize:
        result = normalize_azimuth(result, inplace=inplace)

    return result


@ureg.wraps(ret=None, args=("dimensionless", "rad", None, None), strict=False)
def cos_angle_to_direction(
    cos_theta: np.typing.ArrayLike,
    phi: np.typing.ArrayLike,
    azimuth_convention: AzimuthConvention | str = AzimuthConvention.EAST_RIGHT,
    flip: bool = False,
) -> np.ndarray:
    r"""
    Convert a zenith cosine and azimuth angle pair to a direction.

    Parameters
    ----------
    cos_theta : array-like
        Zenith angle cosine [dimensionless].
        Convention: 1 corresponds to zenith, -1 corresponds to nadir.

    phi : array-like
        Azimuth angle [radian].
        Convention: :math:`2 \pi` corresponds to the X axis.

    azimuth_convention : .AzimuthConvention or str, optional, default: .AzimuthConvention.EAST_RIGHT
        Source azimuth angle convention. If a string is passed, it will be
        converted to a :class:`.AzimuthConvention`.

    flip : bool
        If ``True``, flip the returned direction (points towards the nadir with
        `cos_theta` equal to 1).

    Returns
    -------
    ndarray
        Directions corresponding to the angular parameters [unitless].
    """
    cos_theta = np.atleast_1d(cos_theta).astype(float)
    phi = np.atleast_1d(
        transform_azimuth(
            phi,
            from_convention=azimuth_convention,
            to_convention=AzimuthConvention.EAST_RIGHT,
        )
    )

    sin_theta = np.sqrt(1.0 - np.multiply(cos_theta, cos_theta))
    sin_phi, cos_phi = np.sin(phi), np.cos(phi)

    result = np.vstack((sin_theta * cos_phi, sin_theta * sin_phi, cos_theta)).T
    return result if not flip else -result


@ureg.wraps(ret=None, args=("rad", None, None), strict=False)
def angles_to_direction(
    angles: np.typing.ArrayLike,
    azimuth_convention: AzimuthConvention | str = AzimuthConvention.EAST_RIGHT,
    flip: bool = False,
) -> np.ndarray:
    r"""
    Convert a zenith and azimuth angle pair to a direction unit vector.

    Parameters
    ----------
    angles : array-like
        A sequence of (zenith, azimuth) pairs, where zenith = 0 corresponds to
        +z direction [rad].

    azimuth_convention : .AzimuthConvention or str, optional, default: .AzimuthConvention.EAST_RIGHT
        Source azimuth angle convention. If a string is passed, it will be
        converted to a :class:`.AzimuthConvention`.

    flip : bool, optional, default: False
        If ``True``, flip returned directions (useful in practice to get
        direction vectors pointing towards the local frame origin).

    Returns
    -------
    ndarray
        Directions corresponding to the angular parameters [unitless].
    """
    # Ensure correct shape
    angles = np.atleast_1d(angles).astype(float)
    if angles.ndim < 2:
        angles = angles.reshape((angles.size // 2, 2))
    if angles.ndim > 2 or angles.shape[1] != 2:
        raise ValueError(f"array must be of shape (N, 2), got {angles.shape}")

    # Pre-process zenith
    negative_zenith = angles[:, 0] < 0
    angles[negative_zenith, 0] *= -1
    angles[negative_zenith, 1] += np.pi

    return cos_angle_to_direction(
        np.cos(angles[:, 0]),
        angles[:, 1],
        flip=flip,
        azimuth_convention=azimuth_convention,
    )


@ureg.wraps(ret="rad", args=(None, None, None), strict=False)
def direction_to_angles(
    v: np.typing.ArrayLike,
    azimuth_convention: AzimuthConvention | str = AzimuthConvention.EAST_RIGHT,
    normalize: bool = True,
) -> np.ndarray:
    """
    Convert a cartesian unit vector to a zenith-azimuth pair.

    Parameters
    ----------
    v : array-like
        A sequence of 3-vectors (shape (N, 3)) [unitless].

    azimuth_convention : .AzimuthConvention or str, optional, default: .AzimuthConvention.EAST_RIGHT
        Target azimuth angle convention. If a string is passed, it will be
        converted to a :class:`.AzimuthConvention`.

    normalize : bool, optional, default: True
        If ``True``, normalize azimuth values within the [0, 2π[ interval.

    Returns
    -------
    quantity
        A sequence of 2-vectors containing zenith and azimuth angles, where
        zenith = 0 corresponds to +z direction (shape (N, 2)).
    """
    v = np.atleast_1d(v).astype(float)
    if v.ndim < 2:
        v = v.reshape((v.size // 3, 3))
    if v.ndim > 2 or v.shape[1] != 3:
        raise ValueError(f"array must be of shape (N, 3), got {v.shape}")

    v = v / np.linalg.norm(v, axis=-1).reshape(len(v), 1)
    theta = np.arccos(v[..., 2])
    phi = transform_azimuth(
        np.arctan2(v[..., 1], v[..., 0]),
        from_convention=AzimuthConvention.EAST_RIGHT,
        to_convention=azimuth_convention,
        normalize=normalize,
    )

    return np.vstack((theta, phi)).T


@ureg.wraps(ret=None, args=(None, "rad", "rad", None), strict=False)
def spherical_to_cartesian(r, theta, phi, origin=(0, 0, 0)):
    r"""
    Convert spherical coordinates to cartesian coordinates

    Parameters
    ----------
    r : float
        Radial distance coordinate.

    theta : float
        Zenith angle coordinate [radian].
        Convention: 0 corresponds to zenith, :math:`\pi` corresponds to nadir.

    phi : float
        Azimuth angle coordinate [radian].
        Convention: :math:`2 \pi` corresponds to the X axis.

    origin : array-like
        Shifts the center point of the coordinate system.

    Returns
    -------
    ndarray
        Cartesian coordinates x, y, z as a (3,) array.
    """
    # TODO: Vectorise

    # fmt: off
    if isinstance(r, pint.Quantity):
        return np.array([
            r.magnitude * np.sin(theta) * np.cos(phi) + origin[0],
            r.magnitude * np.sin(theta) * np.sin(phi) + origin[1],
            r.magnitude * np.cos(theta) + origin[2]
        ]) * r.units
    else:
        return np.array([
            r * np.sin(theta) * np.cos(phi) + origin[0],
            r * np.sin(theta) * np.sin(phi) + origin[1],
            r * np.cos(theta) + origin[2]
        ])
    # fmt: on


def angles_in_hplane(
    plane: float,
    theta: np.typing.ArrayLike,
    phi: np.typing.ArrayLike,
    raise_exc: bool = True,
):
    """
    Check that a set of (zenith, azimuth) pairs belong to a given hemisphere
    plane cut.

    Parameters
    ----------
    plane : float
        Plane cut orientation [rad].

    theta : ndarray
        List of zenith angle values with (N,) shape [rad].

    phi : ndarray
        List of azimuth angle values with (N,) shape [rad].

    raise_exc : bool, optional
        If ``True``, raise if not all directions are snapped to the specified
        hemisphere plane cut.

    Returns
    -------
    in_plane_positive, in_plane_negative
        Masks indicating indexes of directions contained in the positive (resp.
        negative) half-plane.

    Raises
    ------
    ValueError
        If not all directions are snapped to the specified hemisphere plane cut.
    """
    # Normalize input parameters
    twopi = 2.0 * np.pi
    phi = np.where(theta >= 0.0, phi % twopi, (phi + np.pi) % twopi)
    theta = np.where(theta >= 0.0, theta, -theta)

    # Assign angle pairs to positive or negative half-plane
    in_plane_positive = np.isclose(plane, phi) | np.isclose(0.0, theta)
    in_plane_negative = np.isclose((plane + np.pi) % twopi, phi) & ~in_plane_positive
    in_plane = in_plane_positive | in_plane_negative

    if raise_exc and not (np.all(in_plane)):
        raise ValueError("found off-plane directions")

    return in_plane_positive, in_plane_negative



from __future__ import annotations

import typing as t
from abc import ABC
from collections.abc import MutableMapping
from pathlib import Path

import attrs
import mitsuba as mi
import pint
import pinttr

from ._core import CanopyElement, biosphere_factory
from ._leaf_cloud import LeafCloud
from ..core import SceneElement, traverse
from ..spectra import Spectrum, spectrum_factory
from ... import validators
from ...attrs import define, documented, get_doc
from ...kernel import SceneParameter, SearchSceneParameter
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg


@define(eq=False, slots=False)
class Tree(CanopyElement, ABC):
    """
    Abstract base class for tree-like canopy elements.
    """

    pass


def _leaf_cloud_converter(value):
    """
    Special converter for the AbstractTree.leaf_cloud field.
    """
    if isinstance(value, MutableMapping):
        value["type"] = "leaf_cloud"
    return biosphere_factory.convert(value)


@define(eq=False, slots=False)
class AbstractTree(Tree):
    """
    A container class for abstract trees in discrete canopies.
    Holds a :class:`.LeafCloud` and the parameters characterizing a cylindrical
    trunk. The entire tree is described in local coordinates and can be placed
    in the scene using :class:`.InstancedCanopyElement`.

    The trunk starts at [0, 0, -0.1] and extends
    to [0, 0, trunk_height]. The trunk extends below ``z=0`` to avoid
    intersection issues at the intersection of the trunk and the ground the tree
    is usually placed on.

    The leaf cloud will by default be offset such that its local coordinate
    origin coincides with the upper end of the trunk. If this is not desired,
    e.g. the leaf cloud is centered around its coordinate origin and the trunk
    should not extend into it, the parameter ``leaf_cloud_extra_offset`` can be
    used to shift the leaf cloud **in addition** to the trunk's extent.
    """

    id: str | None = documented(
        attrs.field(
            default="abstract_tree",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(SceneElement, "id", "doc"),
        type=get_doc(SceneElement, "id", "type"),
        init_type=get_doc(SceneElement, "id", "init_type"),
        default='"abstract_tree"',
    )

    leaf_cloud: LeafCloud | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(_leaf_cloud_converter),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(LeafCloud)
            ),
        ),
        doc="Instanced leaf cloud. Can be specified as a dictionary, which will "
        "be interpreted by :data:`.biosphere_factory`. If the latter case, the "
        '``"type"`` parameter, if omitted, will implicitly be set to '
        '``"leaf_cloud"``.',
        type=".LeafCloud or None",
        init_type=".LeafCloud or dict, optional",
    )

    trunk_height: pint.Quantity = documented(
        pinttr.field(default=1.0 * ureg.m, units=ucc.deferred("length")),
        doc="Trunk height. Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="quantity or float",
        default="1.0 m",
    )

    trunk_radius: pint.Quantity = documented(
        pinttr.field(default=0.1 * ureg.m, units=ucc.deferred("length")),
        doc="Trunk radius. Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="quantity or float, optional",
        default="0.1 m",
    )

    trunk_reflectance: Spectrum = documented(
        attrs.field(
            default=0.5,
            converter=spectrum_factory.converter("reflectance"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("reflectance"),
            ],
        ),
        doc="Reflectance spectrum of the trunk. "
        "Must be a reflectance spectrum (dimensionless).",
        type=".Spectrum",
        init_type=".Spectrum or dict, optional",
        default="0.5",
    )

    leaf_cloud_extra_offset: pint.Quantity = documented(
        pinttr.field(factory=lambda: [0, 0, 0], units=ucc.deferred("length")),
        doc="Additional offset for the leaf cloud. 3-vector. "
        "Unit-enabled field (default: ucc['length'])",
        type="quantity",
        init_type="quantity or array-like, optional",
        default="[0, 0, 0]",
    )

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def bsdf_id(self) -> str:
        return f"bsdf_{self.id}"

    @property
    def _template_bsdfs(self) -> dict:
        objects = {"reflectance": traverse(self.trunk_reflectance)[0].data}

        result = {f"{self.bsdf_id}.type": "diffuse"}

        for obj_key, obj_params in objects.items():
            for key, param in obj_params.items():
                result[f"{self.bsdf_id}.{obj_key}.{key}"] = param

        return {**self.leaf_cloud._template_bsdfs, **result}

    @property
    def _template_shapes(self) -> dict:
        kernel_length = uck.get("length")

        kernel_height = self.trunk_height.m_as(kernel_length)
        kernel_radius = self.trunk_radius.m_as(kernel_length)

        leaf_cloud = self.leaf_cloud.translated(
            [0.0, 0.0, kernel_height] * kernel_length
            + self.leaf_cloud_extra_offset.to(kernel_length)
        )

        bsdf = {"type": "ref", "id": self.bsdf_id}

        result = {
            **leaf_cloud._template_shapes,
            f"trunk_cyl_{self.id}": {
                "type": "cylinder",
                "bsdf": bsdf,
                "radius": kernel_radius,
                "p0": [0, 0, -0.1],
                "p1": [0, 0, kernel_height],
            },
            f"trunk_cap_{self.id}": {
                "type": "disk",
                "bsdf": bsdf,
                "to_world": mi.ScalarTransform4f.scale(kernel_radius)
                @ mi.ScalarTransform4f.translate([0, 0, kernel_height]),
            },
        }

        return result

    @property
    def _params_bsdfs(self) -> dict:
        objects = {"reflectance": traverse(self.trunk_reflectance)[1].data}

        result = {}

        for obj_key, obj_params in objects.items():
            for key, param in obj_params.items():
                # If no lookup strategy is set, we must add one
                if isinstance(param, SceneParameter) and param.search is None:
                    param = attrs.evolve(
                        param,
                        search=SearchSceneParameter(
                            mi.BSDF,
                            self.bsdf_id,
                            parameter_relpath=f"{obj_key}.{key}",
                        ),
                    )

                result[f"{self.bsdf_id}.{obj_key}.{key}"] = param

        for key, param in self.leaf_cloud._params_bsdfs.items():
            result[key] = param

        return result

    @property
    def _params_shapes(self) -> dict:
        return {}


@define(eq=False, slots=False)
class MeshTree(Tree):
    """
    A container class for mesh based tree-like objects in canopies.

    It holds one or more triangulated meshes and corresponding BSDFs, representing
    the tree.

    The mesh will be interpreted in local coordinates and should be used in an
    :class:`.InstancedCanopyElement` to place at arbitrary positions in a scene.
    """

    id: str = documented(
        attrs.field(
            default="mesh_tree",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(SceneElement, "id", "doc"),
        type=get_doc(SceneElement, "id", "type"),
        init_type=get_doc(SceneElement, "id", "init_type"),
        default='"mesh_tree"',
    )

    mesh_tree_elements: list[MeshTree] = documented(
        attrs.field(
            factory=list,
            converter=lambda value: [
                MeshTreeElement.convert(x) for x in pinttr.util.always_iterable(value)
            ]
            if not isinstance(value, dict)
            else [MeshTreeElement.convert(value)],
        ),
        doc="List of :class:`.CanopyElement` defining the canopy. Can be "
        "initialized with a :class:`.InstancedCanopyElement`, which will be "
        "automatically wrapped into a list. Dictionary-based specifications are "
        "allowed as well.",
        type="list of .InstancedCanopyElement",
        init_type="list of (.InstancedCanopyElement | dict)",
        default="[]",
    )

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def _template_bsdfs(self) -> dict:
        result = {}
        for element in self.mesh_tree_elements:
            result.update(element._template_bsdfs)
        return result

    @property
    def _template_shapes(self) -> dict:
        result = {}
        for element in self.mesh_tree_elements:
            result.update(element._template_shapes)
        return result

    @property
    def _params_bsdfs(self) -> dict:
        result = {}
        for element in self.mesh_tree_elements:
            result.update(element._params_bsdfs)
        return result

    @property
    def _params_shapes(self) -> dict:
        return {}


@define(eq=False, slots=False)
class MeshTreeElement:
    """
    Container class for mesh based constituents of tree-like objects in a canopy.
    Holds the filepath for the triangulated mesh and all parameters specifying
    the associated BSDF.

    .. important:: The triangulated mesh must be provided in .ply or .obj format.

    Since mesh definition files cannot carry Pint units, the attribute ``mesh_units``
    lets users provide the unit which their mesh is defined in. Upon kernel dict
    creation the mesh is scaled to match the length unit used in the kernel.
    If ``mesh_units`` is not provided, no scaling of the mesh is performed.

    The :meth:`MeshTreeElement.from_dict` constructor instantiates the class from
    a configuration dictionary.
    """

    id: str | None = documented(
        attrs.field(
            default="mesh_tree_element",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc="User-defined object identifier.",
        type="str, optional",
    )

    mesh_filename: Path = documented(
        attrs.field(
            converter=Path,
            kw_only=True,
        ),
        doc="Path to the triangulated mesh data file. This parameter is required.",
        type=":class:`pathlib.Path`",
        init_type="path-like",
    )

    @mesh_filename.validator
    def _mesh_filename_validator(self, attribute, value):
        validators.path_exists(self, attribute, value)

        if value.suffix not in {".obj", ".ply"}:
            raise ValueError(
                f"While validating {attribute.name}: File extension must be '.obj'"
                f"or '.ply', got '{value.suffix}'"
            )

    mesh_units: pint.Unit | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(ureg.Unit),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(pint.Unit)
            ),
        ),
        doc="Units the mesh was defined in. Used to convert to kernel units. "
        "If unset, the mesh is interpreted as being defined in kernel units.",
        type=":class:`pint.Unit`, optional",
        init_type="str or :class:`pint.Unit`, optional",
    )

    reflectance: Spectrum = documented(
        attrs.field(
            default=0.5,
            converter=spectrum_factory.converter("reflectance"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("reflectance"),
            ],
        ),
        doc="Reflectance of the object. "
        "Must be a reflectance spectrum (dimensionless).",
        type=":class:`.Spectrum`",
        init_type=":class:`.Spectrum` or dict",
        default="0.5",
    )

    transmittance: Spectrum = documented(
        attrs.field(
            default=0.0,
            converter=spectrum_factory.converter("transmittance"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("transmittance"),
            ],
        ),
        doc="Transmittance of the object. "
        "Must be a transmittance spectrum (dimensionless).",
        type=":class:`.Spectrum`",
        init_type=":class:`.Spectrum` or dict",
        default="0.0",
    )

    # --------------------------------------------------------------------------
    #                              Constructors
    # --------------------------------------------------------------------------

    @staticmethod
    def from_dict(d: dict) -> MeshTreeElement:
        """
        Create from a dictionary. This class method will additionally pre-process
        the passed dictionary to merge any field with an associated ``"_units"``
        field into a :class:`pint.Quantity` container.

        Parameters
        ----------
        d : dict
            Configuration dictionary used for initialization.

        Returns
        -------
        :class:`.MeshTreeElement`
            Created object.
        """

        # Pre-process dict: apply units to unit-enabled fields
        d_copy = pinttr.interpret_units(d, ureg=ureg)

        # Perform object creation
        return MeshTreeElement(**d_copy)

    @staticmethod
    def convert(value: t.Any) -> t.Any:
        """
        Object converter method.

        If ``value`` is a dictionary, this method uses :meth:`.from_dict` to
        create an :class:`.MeshTreeElement`. Otherwise, it returns ``value``.

        Parameters
        ----------
        value
            Value to attempt conversion of.

        Returns
        -------
        any type
            Converted value.
        """
        if isinstance(value, dict):
            return MeshTreeElement.from_dict(value)

        return value

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def bsdf_id(self) -> str:
        return f"bsdf_{self.id}"

    @property
    def _template_bsdfs(self) -> dict:
        objects = {
            "reflectance": traverse(self.reflectance)[0].data,
            "transmittance": traverse(self.transmittance)[0].data,
        }

        result = {f"{self.bsdf_id}.type": "bilambertian"}

        for obj_key, obj_params in objects.items():
            for key, param in obj_params.items():
                result[f"{self.bsdf_id}.{obj_key}.{key}"] = param

        return result

    @property
    def _template_shapes(self) -> dict:
        scaling_factor = (
            1.0
            if self.mesh_units is None
            else ureg.convert(1.0, self.mesh_units, uck.get("length"))
        )

        if self.mesh_filename.suffix == ".obj":
            shape_type = "obj"
        elif self.mesh_filename.suffix == ".ply":
            shape_type = "ply"
        else:
            raise ValueError(
                f"unsupported file extension '{self.mesh_filename.suffix}'"
            )

        result = {
            f"{self.id}.type": shape_type,
            f"{self.id}.bsdf.type": "ref",
            f"{self.id}.bsdf.id": self.bsdf_id,
            f"{self.id}.filename": str(self.mesh_filename),
            f"{self.id}.to_world": mi.ScalarTransform4f.scale(scaling_factor),
        }

        return result

    @property
    def _params_bsdfs(self) -> dict:
        objects = {
            "reflectance": traverse(self.reflectance)[1].data,
            "transmittance": traverse(self.transmittance)[1].data,
        }

        result = {}

        for obj_key, obj_params in objects.items():
            for key, param in obj_params.items():
                # If no lookup strategy is set, we must add one
                if isinstance(param, SceneParameter) and param.search is None:
                    param = attrs.evolve(
                        param,
                        search=SearchSceneParameter(
                            mi.BSDF,
                            self.bsdf_id,
                            parameter_relpath=f"{obj_key}.{key}",
                        ),
                    )

                result[f"{self.bsdf_id}.{obj_key}.{key}"] = param

        return result

    @property
    def _params_shapes(self) -> dict:
        return {}



from __future__ import annotations

import typing as t
import warnings

import matplotlib.pyplot as _plt
import numpy as np
import seaborn as sns
import xarray as xr
from cycler import cycler
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colors import Normalize
from matplotlib.figure import Figure
from matplotlib.gridspec import SubplotSpec
from xarray.plot import FacetGrid

from .radprops import CKDAbsorptionDatabase, MonoAbsorptionDatabase


def set_style(rc=None):
    """
    Apply Eradiate Matplotlib style (the Seaborn ``ticks`` style with the
    viridis colormap).

    Parameters
    ----------
    rc : dict, optional
        A Matplotlib rc parameter dictionary to be applied in addition to the
        Eradiate style.
    """
    try:
        import seaborn
    except ModuleNotFoundError:
        warnings.warn(
            "To use Eradiate's Matplotlib style, you must install Seaborn.\n"
            "See instructions on https://seaborn.pydata.org/installing.html."
        )
        raise

    if rc is None:
        rc = {}

    seaborn.set_theme(style="ticks", rc={"image.cmap": "viridis", **rc})


def detect_axes(from_=None):
    """
    Try and extract a :class:`~matplotlib.axes.Axes` list from a data structure.

    Parameters
    ----------
    from_ : matplotlib figure or axes or list of axes or :class:`xarray.plot.FacetGrid`, optional
        Data structure to get an :class:`~matplotlib.axes.Axes` list from.
        If unset, :func:`matplotlib.pyplot.gca()` is used.

    Returns
    -------
    list of matplotlib axes
        Extracted list of :class:`~matplotlib.axes.Axes`.

    Raises
    ------
    TypeError
        If ``from_`` is of unsupported type.
    """
    if from_ is None:
        from_ = _plt.gca()

    if isinstance(from_, Figure):
        return from_.axes

    if isinstance(from_, Axes):
        return [from_]

    if isinstance(from_, FacetGrid):
        return list(from_.axes.flatten())

    if isinstance(from_, list):
        if all([isinstance(x, Axes) for x in from_]):
            return from_

    raise TypeError("unsupported type")


def get_axes_from_facet_grid(facet_grid: FacetGrid, exclude: str = None) -> list[Axes]:
    """
    Extract a flat list of :class:`~matplotlib.axes.Axes` from a
    :class:`~xarray.plot.FacetGrid`.

    Parameters
    ----------
    facet_grid : :class:`xarray.plot.FacetGrid`
        Object to extract a list of :class:`~matplotlib.axes.Axes`.

    exclude : str, optional
        Exclude selected :class:`~matplotlib.axes.Axes` from
        the returned list. Supported values:

        * ``"lower_left"``: exclude lower-left corner.

    Returns
    -------
    list of matplotlib axes
        List of extracted :class:`~matplotlib.axes.Axes` objects.
    """
    index = np.full((facet_grid.axes.size,), True, dtype=bool)
    n_rows, n_cols = facet_grid.axes.shape

    # This filtering system can be extended in the future
    if exclude == "lower_left":
        index[(n_rows - 1) * n_cols] = False

    return list(facet_grid.axes.flatten()[index])


def remove_xylabels(from_=None) -> None:
    """
    Remove x- and y-axis labels from ``from_``
    (processed by :func:`detect_axes`).

    Parameters
    ----------
    from_ : matplotlib figure or axes or list of axes or :class:`xarray.plot.FacetGrid`, optional
        Data structure to get an :class:`~matplotlib.axes.Axes` list from.
        If unset, :func:`matplotlib.pyplot.gca()` is used.

    See Also
    --------
    :func:`detect_axes`
    """
    for ax in detect_axes(from_):
        ax.set_xlabel("")
        ax.set_ylabel("")


def remove_xyticks(from_=None) -> None:
    """
    Remove x and y axis tick labels from ``from_``
    (processed by :func:`detect_axes`).

    Parameters
    ----------
    from_ : matplotlib figure or axes or list of axes or :class:`xarray.plot.FacetGrid`, optional
        Data structure to get an :class:`~matplotlib.axes.Axes` list from.
        If unset, :func:`matplotlib.pyplot.gca()` is used.

    See Also
    --------
    :func:`detect_axes`
    """
    for ax in detect_axes(from_):
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)


def make_ticks(num_ticks: int, limits: t.Sequence[float]):
    """
    Generate ticks and their respective tickmarks.

    Parameters
    ----------
    num_ticks : int
        Number of ticks to generate, including the limits
        of the given range

    limits : tuple[float, float]
        List of two values, limiting the ticks inclusive

    Returns
    ------
    steps : list of float
        Values for the ticks.

    marks : list of str
        Tick values converted to degrees as string tickmarks.
    """

    step_width = float(limits[1] - limits[0]) / (num_ticks - 1)

    steps = [limits[0] + step_width * i for i in range(num_ticks)]
    marks = [f"{i / np.pi * 180}°" for i in steps]

    return steps, marks


# -- Data dashboard definitions ------------------------------------------------


class PiecewiseNorm(Normalize):
    def __init__(self, levels, clip=False):
        # input levels
        self._levels = np.sort(levels)
        # corresponding normalized values between 0 and 1
        self._normed = np.linspace(0, 1, len(levels))
        Normalize.__init__(self, None, None, clip)

    def __call__(self, value, clip=None):
        # linearly interpolate to get the normalized value
        return np.ma.masked_array(np.interp(value, self._levels, self._normed))

    def inverse(self, value):
        return 1.0 - self.__call__(value)


def _chunk_array(array, len_chunk):
    n_chunks = max(1, round(len(array) / len_chunk))
    return np.array_split(array, n_chunks)


def dashboard_particle_dataset(
    ds: xr.Dataset,
    title: str | None = None,
    phase_wavelength_chunk_size: int = 5,
    palette="muted",
):
    """
    Create a summary dashboard plot for a particle dataset. The generated plot
    displays the extinction coefficient, the albedo and phase function.

    Parameters
    ----------
    ds : Dataset
        Particle (aerosol, cloud, etc.) dataset.

    title : str, optional
        A title for the plot.

    phase_wavelength_chunk_size : int, default: 5
        The produced dashboard shows the phase function clustered in chunks on
        a number of polar plots set by this chunk size.

    palette : str, list, dict, or matplotlib.colors.Colormap, optional
        A Seaborn palette specification used to colour the phase function line
        plots.
    """
    df = ds.phase.isel(i=0, j=0, drop=True).to_dataframe()[["phase"]]
    df = df.sort_values(["w", "mu"]).reset_index()
    df["theta"] = np.arccos(df["mu"])
    phase_min = df["phase"].min()
    phase_max = df["phase"].max()
    # print(f"{phase_min = }, {phase_max = }")

    w_chunks = [
        (chunk.min(), chunk.max())
        for chunk in _chunk_array(np.unique(df["w"]), phase_wavelength_chunk_size)
    ]
    naxs_phase = len(w_chunks)
    ncols = 2
    nrows_phase = naxs_phase // ncols + (1 if naxs_phase % ncols else 0)
    height_ratios = [0.5, 0.5] + [1 for _ in range(nrows_phase)]

    fig, axs = plt.subplot_mosaic(
        [
            ["sigma_t", "sigma_t"],
            ["albedo", "albedo"],
        ]
        + [[f"phase_{i * 2}", f"phase_{i * 2 + 1}"] for i in range(nrows_phase)],
        height_ratios=height_ratios,
        figsize=(4 * ncols, 4 * sum(height_ratios)),
        subplot_kw={"projection": "polar"},
        layout="constrained",
    )

    for i_var, var in enumerate(["sigma_t", "albedo"]):
        gs = axs[var].get_gridspec()
        start = i_var * ncols
        stop = start + 1
        axs[var].remove()
        axs[var] = plt.subplot(SubplotSpec(gs, start, stop), projection=None)
        ds[var].plot(ax=axs[var], ls=":", marker=".")

    for i_phase, (wmin, wmax) in enumerate(w_chunks):
        _df = df.where((df["w"] >= wmin) & (df["w"] <= wmax)).dropna()
        ax = axs[f"phase_{i_phase}"]
        sns.lineplot(
            ax=ax,
            data=_df,
            x="theta",
            y="phase",
            hue="w",
            hue_norm=PiecewiseNorm(levels=_df["w"].values),
            legend="full",
            palette=palette,
        )
        ax.set_rlim([phase_min * 0.5, phase_max * 5.0])
        ax.set_rscale("log")
        ax.set_thetagrids(np.arange(0, 181, 45))
        ax.set_thetamax(180)
        ax.set_xlabel("")
        ax.set_ylabel("")

        # Deal with legend entries
        lgd_title = "wavelength"  # default value, overridden by metadata if any
        for field in ["long_name", "standard_name"]:
            if field in ds.w.attrs:
                lgd_title = ds.w.attrs[field]
                break

        if "units" in ds.w.attrs:
            lgd_title += f" [{ds.w.units}]"

        lgd = ax.legend(
            loc="upper center",
            ncol=3,
            bbox_to_anchor=(0.5, 0.1),
            title=lgd_title,
        )
        new_labels = [f"{w:.3g}" for w in np.unique(_df["w"].values)]
        for i, _ in enumerate(lgd.get_texts()):
            lgd.get_texts()[i].set_text(new_labels[i])

    # Clear empty axes
    for i in range(i_phase + 1, nrows_phase * 2):
        ax = axs[f"phase_{i}"]
        ax.remove()

    if title:
        fig.suptitle(title)

    return fig, axs


def absorption_database_spectral_coverage_mono(
    db: MonoAbsorptionDatabase, wrange=None, title=None
):
    df_index = db._index

    fig, ax = plt.subplots(1, 1, figsize=(6, 1.5))

    color_cycle = cycler(color=sns.color_palette())

    for i, (filename, color) in enumerate(zip(df_index["filename"], color_cycle())):
        color = color["color"]
        spectral_coverage = db.spectral_coverage.loc[filename]
        w = spectral_coverage.index.values
        sns.histplot(x=w, color=color, ax=ax)

    ax.set_xlabel("Wavelength [nm]")
    if title:
        ax.set_title(title)
    sns.despine(offset=10)

    return fig, ax


def absorption_database_spectral_coverage_ckd(
    db: CKDAbsorptionDatabase, wrange=None, title=None
):
    df_index = db._index
    wmins = df_index["wl_min [nm]"]
    wmaxs = df_index["wl_max [nm]"]

    if wrange is not None:
        select = (wmins >= wrange[0]) & (wmaxs <= wrange[1])
        df_index = df_index.loc[select]

    wmins = df_index["wl_min [nm]"]
    wmaxs = df_index["wl_max [nm]"]

    fig, ax = plt.subplots(1, 1, figsize=(6, 1))

    color_cycle = cycler(color=sns.color_palette())

    for i, (filename, color) in enumerate(zip(df_index["filename"], color_cycle())):
        color = color["color"]
        spectral_coverage = db.spectral_coverage.loc[filename]

        for _, (wmin, wmax) in spectral_coverage.iterrows():
            ax.fill_between([wmin, wmax], 0, 1, color=color, alpha=0.5)

        if i == 0:
            w = wmins.min()
            ax.scatter(w, 0.5, color=color, marker=8)
            ax.annotate(
                f"{w:4.2f}",
                (w, 0.5),
                color=color,
                textcoords="offset points",
                xytext=(-6, 0),
                ha="right",
                va="center",
            )

    else:
        w = wmaxs.max()
        ax.scatter(w, 0.5, color=color, marker=9)
        ax.annotate(
            f"{w:4.2f}",
            (w, 0.5),
            color=color,
            textcoords="offset points",
            xytext=(6, 0),
            ha="left",
            va="center",
        )

    ax.set_xlabel("Wavelength [nm]")
    ax.set_yticks([])
    if title:
        ax.set_title(title)
    sns.despine(ax=ax, top=True, left=True, right=True, bottom=False)

    return fig, ax



from rich.console import Console

console = Console(color_system=None)
error_console = Console(stderr=True, color_system=None)


def section(title, newline=True):
    if newline:
        console.print()
    console.rule("── " + title, align="left")
    console.print()


def message(text):
    console.print(text)


def warning(text):
    error_console.print(text)



import drjit as dr
import mitsuba as mi
import numpy as np
import xarray as xr


def _sph_to_dir(theta, phi):
    """Map spherical to Euclidean coordinates"""
    st, ct = dr.sincos(theta)
    sp, cp = dr.sincos(phi)
    return mi.Vector3f(cp * st, sp * st, ct)


def _eval_bsdf_impl(plugin, theta_os, phi_os, theta_i, phi_i):
    theta_ov, phi_ov = dr.meshgrid(theta_os, phi_os)
    wos = _sph_to_dir(theta_ov, phi_ov)

    # Evaluate BSDF
    si = dr.zeros(mi.SurfaceInteraction3f)
    si.wi = _sph_to_dir(theta_i, phi_i)
    values = plugin.eval(mi.BSDFContext(), si, wos)
    return np.reshape(values, (len(phi_os), len(theta_os))).T


def eval_bsdf(plugin, theta_os, phi_os, theta_is, phi_is) -> xr.Dataset:
    # Requires a JIT variant
    if mi.Float == mi.ScalarFloat:
        raise RuntimeError("A JIT-compiled variant is required")

    result = [
        [
            _eval_bsdf_impl(plugin, theta_os, phi_os, theta_i, phi_i)
            for theta_i in theta_is
        ]
        for phi_i in phi_is
    ]

    return xr.Dataset(
        {
            "bsdf": (
                ("theta_o", "phi_o", "theta_i", "phi_i"),
                np.transpose(result, (2, 3, 1, 0)),
                {"units": "sr^-1"},
            )
        },
        coords={
            "theta_o": ("theta_o", theta_os, {"units": "rad"}),
            "phi_o": ("phi_o", phi_os, {"units": "rad"}),
            "theta_i": ("theta_i", theta_is, {"units": "rad"}),
            "phi_i": ("phi_i", phi_is, {"units": "rad"}),
        },
    )



"""The Eradiate radiative transfer simulation software package."""

import importlib
import os

from ._version import _version

__version__ = _version  #: Eradiate version string.

# -- Lazy imports ------------------------------------------------------

import lazy_loader  # noqa: E402

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

del lazy_loader

if not os.environ.get("EAGER_IMPORT"):
    # This performs kernel checks. If eager imports are activated, this import
    # will occur automatically, and it is therefore not necessary to repeat it.
    importlib.import_module("eradiate.kernel._versions")



"""
This module exposes components dedicated to the global configuration of
Eradiate.
"""

import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

del lazy_loader



from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod
from functools import singledispatchmethod

import attrs
import pint

from ..core import NodeSceneElement
from ..._factory import Factory
from ...attrs import define, documented
from ...spectral.index import (
    CKDSpectralIndex,
    MonoSpectralIndex,
    SpectralIndex,
)
from ...units import PhysicalQuantity


class SpectrumFactory(Factory):
    def converter(self, quantity: str | PhysicalQuantity) -> t.Callable[[t.Any], t.Any]:
        """
        Generate a converter wrapping :meth:`SpectrumFactory.convert` to
        handle defaults for shortened spectrum definitions. The produced
        converter processes a parameter ``value`` as follows:

        * if ``value`` is an int, a float or a :class:`pint.Quantity`, the
          converter calls itself using a dictionary
          ``{"type": "uniform", "quantity": quantity, "value": value}``;
        * if ``value`` is a dictionary, it adds a ``"quantity": quantity`` entry
          for the following values of the ``"type"`` entry:

          * ``"uniform"``;
          * ``"interpolated"``;
          * ``"multi_delta"``;

        * otherwise, it forwards ``value`` to :meth:`.SpectrumFactory.convert`.

        Parameters
        ----------
        quantity : str or .PhysicalQuantity
            Quantity specifier (converted by :class:`.PhysicalQuantity`).
            See :meth:`.PhysicalQuantity.spectrum` for suitable values.

        Returns
        -------
        callable
            Generated converter.
        """

        def f(value):
            if isinstance(value, (int, float, pint.Quantity)):
                # Convert unitless values with dict wrapping and recursive call
                return f({"type": "uniform", "quantity": quantity, "value": value})

            if isinstance(value, dict):
                # If generic spectrum is requested without a specified
                # "quantity" field, add pre-configured quantity and attempt
                # conversion using regular conversion protocol
                try:
                    if (
                        value["type"] in {"uniform", "interpolated", "multi_delta"}
                        and "quantity" not in value
                    ):
                        return self.convert({**value, "quantity": quantity})
                except KeyError:
                    # Note: A missing "type" field will also run this case, and
                    # the ill-formed dict will be correctly reported upon
                    # regular conversion
                    pass

            if isinstance(value, Spectrum) and value.quantity is None:
                # If quantity is unspecified, apply the expected quantity
                return self.convert(attrs.evolve(value, quantity=quantity))

            # Regular conversion (happens if value is neither int, float nor
            # dict without "quantity" field)
            return self.convert(value)

        return f


spectrum_factory = SpectrumFactory()
spectrum_factory.register_lazy_batch(
    [
        (
            "_air_scattering_coefficient.AirScatteringCoefficientSpectrum",
            "air_scattering_coefficient",
            {},
        ),
        (
            "_interpolated.InterpolatedSpectrum",
            "interpolated",
            {},
        ),
        (
            "_solar_irradiance.SolarIrradianceSpectrum",
            "solar_irradiance",
            {},
        ),
        (
            "_uniform.UniformSpectrum",
            "uniform",
            {},
        ),
    ],
    cls_prefix="eradiate.scenes.spectra",
)


@define(eq=False, slots=False)
class Spectrum(NodeSceneElement, ABC):
    """
    Spectrum interface.

    Notes
    -----
    * This class is to be used as a mixin.
    * Subclasses must implement :meth:`eval_mono`, :meth:`eval_ckd` and
      :meth:`integral`.

    Warnings
    --------
    Leaving the ``quantity`` field unset is discouraged and permitted only to
    allow for more convenience in some parts of the code (*e.g.* to increase the
    flexibility of spectrum conversion protocols).
    """

    quantity: PhysicalQuantity | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(PhysicalQuantity),
            repr=lambda x: "None" if x is None else x.value.upper(),
        ),
        doc="Physical quantity which the spectrum represents. The specified "
        "quantity must be one which varies with wavelength. "
        "See :meth:`.PhysicalQuantity.spectrum` for allowed values.\n"
        "\n"
        "Child classes should implement value units validation and conversion "
        "based on ``quantity``.",
        type=".PhysicalQuantity or None",
        init_type=".PhysicalQuantity or str, optional",
        default="None",
    )

    @quantity.validator
    def _quantity_validator(self, attribute, value):
        if value is None:
            return

        if value not in PhysicalQuantity.spectrum():
            raise ValueError(
                f"while validating {attribute.name}: "
                f"got value '{value}', expected one of "
                f"{str(PhysicalQuantity.spectrum())}"
            )

    @singledispatchmethod
    def eval(self, si: SpectralIndex) -> pint.Quantity:
        """
        Evaluate spectrum at a given spectral index.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        Returns
        -------
        value : quantity
            Evaluated spectrum.

        Notes
        -----
        This method dispatches evaluation to specialized methods depending
        on the spectral index type.
        """
        raise NotImplementedError

    @eval.register(MonoSpectralIndex)
    def _(self, si) -> pint.Quantity:
        return self.eval_mono(w=si.w)

    @eval.register(CKDSpectralIndex)
    def _(self, si) -> pint.Quantity:
        return self.eval_ckd(w=si.w, g=si.g)

    @abstractmethod
    def eval_mono(self, w: pint.Quantity) -> pint.Quantity:
        """
        Evaluate spectrum in monochromatic modes.

        Parameters
        ----------
        w : quantity
            Wavelength.

        Returns
        -------
        value : quantity
            Evaluated spectrum as an array with the same shape as ``w``.
        """
        raise NotImplementedError

    @abstractmethod
    def eval_ckd(self, w: pint.Quantity, g: float) -> pint.Quantity:
        """
        Evaluate spectrum in CKD modes.

        Parameters
        ----------
        w : quantity
            Spectral bin center wavelength.

        g : float
            Absorption coefficient cumulative probability.

        Returns
        -------
        value : quantity
            Evaluated spectrum as an array with shape ``w``.

        Notes
        -----
        It is assumed that ``w`` and ``g`` have the same shape.
        In CKD mode, it is assumed that all spectra—except that of the
        absorption coefficient—are uniform over the spectral bin. These
        spectra are evaluated at the spectral bin center wavelength.
        """
        pass

    def integral(self, wmin: pint.Quantity, wmax: pint.Quantity) -> pint.Quantity:
        """
        Compute the integral of the spectrum on a given interval.

        Parameters
        ----------
        wmin : quantity
            Integration interval's lower bound.

        wmax : quantity
            Integration interval's upper bound.

        Returns
        -------
        value : quantity or ndarray
            Computed integral value, in units consistent with the specified
            quantity.

        Warnings
        --------
        If the ``quantity`` field is unset (*i.e.* left to its default value
        ``None``), the output of this method will inherit the units of value
        fields (the actual policy depends on the implementation, and unitless
        values are intepreted as dimnesionless). Note that leaving the
        ``quantity`` field unset is discouraged.
        """
        raise NotImplementedError



from .ckd_quad import CKDQuadConfig as CKDQuadConfig
from .ckd_quad import CKDQuadPolicy as CKDQuadPolicy
from .grid import CKDSpectralGrid as CKDSpectralGrid
from .grid import MonoSpectralGrid as MonoSpectralGrid
from .grid import SpectralGrid as SpectralGrid
from .index import CKDSpectralIndex as CKDSpectralIndex
from .index import MonoSpectralIndex as MonoSpectralIndex
from .index import SpectralIndex as SpectralIndex
from .response import BandSRF as BandSRF
from .response import DeltaSRF as DeltaSRF
from .response import SpectralResponseFunction as SpectralResponseFunction
from .response import UniformSRF as UniformSRF



from __future__ import annotations

import attrs

from ._core import BSDF
from ...kernel import SceneParameter


@attrs.define(eq=False, slots=False)
class BlackBSDF(BSDF):
    """
    Black BSDF [``black``].

    This BSDF models a perfectly absorbing surface. It is equivalent to a
    :class:`.DiffuseBSDF` with zero reflectance.
    """

    @property
    def template(self) -> dict:
        # Inherit docstring

        result = {
            "type": "diffuse",
            "reflectance": {"type": "uniform", "value": 0.0},
        }

        if self.id is not None:
            result["id"] = self.id

        return result

    @property
    def params(self) -> dict[str, SceneParameter]:
        # Inherit docstring
        return {}



from __future__ import annotations

from typing import Literal

import numpy as np
import pandas as pd
import xarray as xr
from tqdm.auto import tqdm

from ._core import AbstractHeterogeneousAtmosphere
from ._heterogeneous import HeterogeneousAtmosphere
from ...cfconventions import ATTRIBUTES
from ...spectral import (
    CKDQuadConfig,
    CKDSpectralGrid,
    MonoSpectralGrid,
    MonoSpectralIndex,
)
from ...units import symbol
from ...units import unit_context_config as ucc


def eval_transmittance_mono(
    atmosphere: AbstractHeterogeneousAtmosphere,
    spectral_grid: MonoSpectralGrid,
    interaction: Literal["extinction", "absorption", "scattering"] = "extinction",
) -> xr.DataArray:
    """
    Compute the total column transmittance of an atmosphere instance in
    monochromatic mode.

    Parameters
    ----------
    atmosphere : .AbstractHeterogeneousAtmosphere
        1D atmosphere for which transmittance is evaluated.

    spectral_grid : .MonoSpectralGrid
        The spectral grid on which transmittance is evaluated.

    interaction : {"extinction", "absorption", "scattering"}, optional, default: "extinction"
        The interaction type for which transmittance is evaluated.

    Returns
    -------
    transmittance : DataArray
    """
    wavelengths = spectral_grid.wavelengths
    transmittance = np.full(wavelengths.size, np.nan)

    for i, w in tqdm(
        list(enumerate(spectral_grid.wavelengths)),
        desc="Evaluating transmittance (mono)",
    ):
        si = MonoSpectralIndex(w=w)
        transmittance[i] = atmosphere.eval_transmittance(si=si, interaction=interaction)

    w_u = ucc.get("wavelength")
    result = xr.DataArray(
        transmittance,
        dims=["w"],
        coords={
            "w": (
                "w",
                wavelengths.m_as(w_u),
                {**ATTRIBUTES["radiation_wavelength"], **{"units": symbol(w_u)}},
            )
        },
        attrs={"units": "1", "long_name": "transmittance"},
    )
    return result


def eval_transmittance_ckd(
    atmosphere: AbstractHeterogeneousAtmosphere,
    spectral_grid: CKDSpectralGrid,
    ckd_quad_config: CKDQuadConfig | None = None,
    interaction: Literal["extinction", "absorption", "scattering"] = "extinction",
) -> xr.DataArray:
    """
    Compute the total column transmittance of an atmosphere instance in CKD mode.

    Parameters
    ----------
    atmosphere : .AbstractHeterogeneousAtmosphere
        1D atmosphere for which transmittance is evaluated.

    spectral_grid : .CKDSpectralGrid
        The spectral grid on which transmittance is evaluated.

    ckd_quad_config : CKDQuadConfig, optional
        CKD quadrature rule generation configuration. If unset, a conservative
        default is used (Gauss-Legendre, fixed, with 16 g-points).

    interaction : {"extinction", "absorption", "scattering"}, optional, default: "extinction"
        The interaction type for which transmittance is evaluated.

    Returns
    -------
    transmittance : DataArray
    """
    ckd_quad_config_default = {
        "type": "gauss_legendre",
        "ng_max": 16,
        "policy": "fixed",
    }
    if ckd_quad_config is None:
        ckd_quad_config = ckd_quad_config_default

    if isinstance(ckd_quad_config, dict):
        ckd_quad_config = {**ckd_quad_config_default, **ckd_quad_config}
        ckd_quad_config = CKDQuadConfig(**ckd_quad_config)

    # Check if a molecular absorption database is available
    abs_db = None
    if isinstance(atmosphere, HeterogeneousAtmosphere):
        abs_db = atmosphere.absorption_data
    else:
        pass

    transmittance = {}
    sis = list(spectral_grid.walk_indices(ckd_quad_config, abs_db))

    for si in tqdm(sis, desc="Evaluating transmittance (CKD)"):
        transmittance[si.as_hashable] = atmosphere.eval_transmittance(
            si=si, interaction=interaction
        ).m
    index = pd.MultiIndex.from_tuples(transmittance.keys(), names=("w", "g"))
    da: xr.DataArray = (
        pd.DataFrame(data={"transmittance": list(transmittance.values())}, index=index)
        .to_xarray()
        .transmittance
    )

    # For each bin, compute quadrature and store the result
    result = xr.full_like(da, np.nan).isel(g=0, drop=True)
    w_u = ucc.get("wavelength")

    for i, (w, quad) in enumerate(spectral_grid.walk_quads(ckd_quad_config)):
        values_at_nodes = da.sel(w=w.m).values

        # Rationale: Avoid using xarray's indexing in this loop for
        # performance reasons (wrong data indexing method will result in
        # 10x+ speed reduction)
        result.values[i] = quad.integrate(values_at_nodes[:], interval=(0.0, 1.0))

    result = xr.DataArray(
        result,
        dims=["w"],
        coords={
            "w": (
                "w",
                spectral_grid.wavelengths.m_as(w_u),
                {**ATTRIBUTES["radiation_wavelength"], **{"units": symbol(w_u)}},
            )
        },
        attrs={"units": "1", "long_name": "transmittance"},
    )

    return result



from __future__ import annotations

import typing as t

import attrs

from ._core import EarthObservationExperiment
from ._helpers import surface_converter
from .. import validators
from ..attrs import AUTO, define, documented
from ..scenes.biosphere import Canopy, biosphere_factory
from ..scenes.bsdfs import LambertianBSDF
from ..scenes.core import SceneElement
from ..scenes.integrators import VolPathIntegrator
from ..scenes.measure import Measure
from ..scenes.shapes import RectangleShape
from ..scenes.surface import BasicSurface


@define
class CanopyExperiment(EarthObservationExperiment):
    """
    Simulate radiation in a scene with an explicit canopy and no atmosphere.
    This experiment assumes that the surface is plane and accounts for ground
    unit cell padding.

    Notes
    -----
    A post-initialization step will constrain the measure setup if a
    distant measure is used and no target is defined:

    * if a canopy is defined, the target will be set to the top of the canopy
      unit cell (*i.e.* without its padding);
    * if no canopy is defined, the target will be set to [0, 0, 0].
    """

    canopy: Canopy | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(biosphere_factory.convert),
            validator=attrs.validators.optional(attrs.validators.instance_of(Canopy)),
        ),
        doc="Canopy specification. "
        "This parameter can be specified as a dictionary which will be "
        "interpreted by :data:`.biosphere_factory`.",
        type=":class:`.Canopy` or None",
        init_type=":class:`.Canopy` or dict or None",
        default="None",
    )

    padding: int = documented(
        attrs.field(default=0, converter=int, validator=validators.is_positive),
        doc="Padding level. The scene will be padded with copies to account for "
        "adjacency effects. This, in practice, has effects similar to "
        "making the scene periodic."
        "A value of 0 will yield only the defined scene. A value of 1 "
        "will add one copy in every direction, yielding a 3×3 patch. A "
        "value of 2 will yield a 5×5 patch, etc. The optimal padding level "
        "depends on the scene.",
        type="int",
        default="0",
    )

    surface: None | BasicSurface = documented(
        attrs.field(
            factory=lambda: LambertianBSDF(),
            converter=attrs.converters.optional(surface_converter),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(BasicSurface)
            ),
        ),
        doc="Surface specification. If set to ``None``, no surface will be "
        "added. This parameter can be specified as a dictionary which will be "
        "interpreted by :data:`.surface_factory` and :data:`.bsdf_factory`. "
        "**If relevant, the surface size will be adjusted automatically upon "
        "kernel. dictionary generation.**",
        type=".BasicSurface or None",
        init_type=".BasicSurface or .BSDF or dict, optional",
        default=":class:`BasicSurface(bsdf=LambertianBSDF()) <.BasicSurface>`",
    )

    def __attrs_post_init__(self):
        self._normalize_spectral()
        self._normalize_measures()
        self._normalize_integrator()

    def _normalize_measures(self) -> None:
        """
        Ensure that distant measure targets are set to appropriate values.
        Processed measures will have its ray target and origin parameters
        overridden if relevant.
        """
        for measure in self.measures:
            # Override ray target location if relevant
            if (
                measure.is_distant() and measure.target is None
            ):  # No target specified: add one
                if self.canopy is None:  # No canopy: target origin point
                    measure.target = {"type": "point", "xyz": [0, 0, 0]}

                else:  # Canopy: target top of canopy
                    measure.target = {
                        "type": "rectangle",
                        "xmin": -0.5 * self.canopy.size[0],
                        "xmax": 0.5 * self.canopy.size[0],
                        "ymin": -0.5 * self.canopy.size[1],
                        "ymax": 0.5 * self.canopy.size[1],
                        "z": self.canopy.size[2],
                    }

    def _normalize_integrator(self) -> None:
        """
        Ensures that the integrator is compatible with the atmosphere and geometry.
        """

        if self.integrator is AUTO:
            self.integrator = VolPathIntegrator()

    def _dataset_metadata(self, measure: Measure) -> dict[str, str]:
        result = super()._dataset_metadata(measure)

        if measure.is_distant():
            result["title"] = "Top-of-canopy simulation results"

        return result

    def _context_kwargs(self) -> dict[str, t.Any]:
        return {}

    @property
    def scene_objects(self) -> dict[str, SceneElement]:
        # Inherit docstring

        objects = {}

        # Process canopy and surface
        if self.canopy is not None:
            scene_width = max(self.canopy.size[:2])

            if self.padding > 0:  # Add extra instances if padding is requested
                scene_width *= 2.0 * self.padding + 1.0
                objects["canopy"] = self.canopy.padded_copy(self.padding)
            else:
                objects["canopy"] = self.canopy

            if self.surface is not None:  # Adjust surface to match canopy
                objects["surface"] = attrs.evolve(
                    self.surface,
                    shape=RectangleShape(center=[0, 0, 0], edges=scene_width),
                )

        else:
            if self.surface is not None:  # Leave surface unchanged
                objects["surface"] = self.surface

        objects.update(
            {
                "illumination": self.illumination,
                **{measure.id: measure for measure in self.measures},
                "integrator": self.integrator,
            }
        )

        return objects



import numpy as np

from ...experiments import AtmosphereExperiment
from ...units import unit_registry as ureg


# TODO : This probably needs to go somewhere else or get called by fixtures
def absorption_database_error_handler_config():
    """
    Error handler configuration for absorption coefficient interpolation.

    Notes
    -----
    This configuration is chosen to ignore all interpolation issues (except
    bounds error along the mole fraction dimension) because warnings are
    captured by pytest which will raise.
    Ignoring the bounds on pressure and temperature is safe because
    out-of-bounds values usually correspond to locations in the atmosphere
    that are so high that the contribution to the absorption coefficient
    are negligible at these heights.
    The bounds error for the 'x' (mole fraction) coordinate is considered
    fatal.
    """
    return {
        "p": {"missing": "raise", "scalar": "raise", "bounds": "ignore"},
        "t": {"missing": "raise", "scalar": "raise", "bounds": "ignore"},
        "x": {"missing": "ignore", "scalar": "ignore", "bounds": "raise"},
    }


def create_rpv_afgl1986_continental_brfpp(absorption_database_error_handler_config):
    r"""
    *RPV AFGL1986 Aerosol*

    *Scene setup*

    - Geometry: 1D plane-parallel
    - Atmosphere: Molecular atmosphere using the AFGL 1986 (U.S. Standard) profile
    - Aerosol layer: Uniform layer ranging from 1 km to 2 km, with AOT at 550 nm = 0.5; aerosol dataset ``govaerts_2021-continental``
    - Surface: RPV BSDF with k = 0.95, g = -0.1, and ρ₀ = 0.027685 (vegetated canopy)
    - Illumination: Directional illumination with a zenith angle of 20°
    - Sensor: Multi-distant measure covering the principal plane, from -75° to 75° with 2° increments, delta SRF positioned at λ = 550 nm
    """

    return AtmosphereExperiment(
        surface={"type": "rpv", "k": 0.95, "g": -0.1, "rho_0": 0.027685},
        illumination={
            "type": "directional",
            "zenith": 20 * ureg.deg,
            "irradiance": 20.0,
        },
        measures={
            "type": "mdistant",
            "construct": "hplane",
            "azimuth": 0.0,
            "zeniths": np.arange(-75.0, 75.01, 2.0),
            "srf": {"type": "delta", "wavelengths": 550 * ureg.nm},
        },
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": {
                "type": "molecular",
                "thermoprops": {
                    "identifier": "afgl_1986-us_standard",
                    "z": np.arange(0, 120.1, 0.1) * ureg.km,
                    "additional_molecules": False,
                },
                "absorption_data": "monotropa",
                "error_handler_config": absorption_database_error_handler_config,
            },
            "particle_layers": {
                "type": "particle_layer",
                "bottom": 1 * ureg.km,
                "top": 2 * ureg.km,
                "tau_ref": 0.5,
                "dataset": "govaerts_2021-continental",
            },
        },
        integrator={"type": "piecewise_volpath", "moment": True},
    )


def create_rpv_afgl1986_brfpp(absorption_database_error_handler_config):
    r"""
    *RPV AFGL1986*

    *Scene setup*

    - Geometry: 1D plane-parallel
    - Atmosphere: Molecular atmosphere using the AFGL 1986 (U.S. Standard) profile
    - Surface: RPV BSDF with k = 0.95, g = -0.1, and ρ₀ = 0.027685 (vegetated canopy)
    - Illumination: Directional illumination with a zenith angle of 20°
    - Sensor: Multi-distant measure covering the principal plane, from -75° to 75° with 2° increments, delta SRF positioned at λ = 550 nm
    """
    return AtmosphereExperiment(
        surface={"type": "rpv", "k": 0.95, "g": -0.1, "rho_0": 0.027685},
        illumination={
            "type": "directional",
            "zenith": 20 * ureg.deg,
            "irradiance": 20.0,
        },
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "azimuth": 0.0,
                "zeniths": np.arange(-75.0, 75.01, 2.0),
                "srf": {"type": "delta", "wavelengths": 550 * ureg.nm},
            }
        ],
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": {
                "type": "molecular",
                "thermoprops": {
                    "identifier": "afgl_1986-us_standard",
                    "z": np.linspace(0, 120, 61) * ureg.km,
                    "additional_molecules": False,
                },
                "absorption_data": "monotropa",
                "error_handler_config": absorption_database_error_handler_config,
            },
        },
        ckd_quad_config={"ng_max": 16},
        integrator={"type": "piecewise_volpath", "moment": True},
    )



import joseki  # noqa: F401  # Put import at top to mitigate undesired log output
import typer

app = typer.Typer()


@app.command()
def main():
    """
    Display information useful for debugging.
    """
    import eradiate.util.sys_info

    from ._console import message, section, warning

    warnings = eradiate.kernel.check_kernel()

    if warnings:
        section("Warnings", newline=False)
    for w in warnings:
        warning("• " + w)

    sys_info = eradiate.util.sys_info.show()
    section("System")
    message(f"CPU: {sys_info['cpu_info']}")
    message(f"OS: {sys_info['os']}")
    message(f"Python: {sys_info['python']}")

    section("Versions")
    message(f"• eradiate {eradiate.__version__}")
    message(f"• drjit {sys_info['drjit_version']}")
    message(f"• mitsuba {sys_info['mitsuba_version']}")
    message(f"• eradiate-mitsuba {sys_info['eradiate_mitsuba_version']}")
    message(f"• numpy {sys_info['numpy']}")
    message(f"• scipy {sys_info['scipy']}")
    message(f"• xarray {sys_info['xarray']}")

    try:
        import mitsuba as mi

        section("Available Mitsuba variants")
        message("\n".join([f"• {variant}" for variant in mi.variants()]))
    except ImportError:
        pass

    section("Configuration")
    for var in ["SOURCE_DIR"]:
        value = getattr(eradiate.config, var)
        var_repr = str(value)
        message(f"• ERADIATE_{var.upper()}: {var_repr}")

    loaded_settings_files = list(eradiate.config.settings._loaded_files)
    if loaded_settings_files:
        message("• Loaded settings files:")
        for fname in loaded_settings_files:
            message(f"  • {fname}")
    else:
        message("• Loaded setting files: <none>")


__doc__ = main.__doc__



from __future__ import annotations

import attrs
import mitsuba as mi

from ._core import BSDF
from ..core import traverse
from ..spectra import Spectrum, spectrum_factory
from ... import validators
from ...attrs import define, documented
from ...kernel import SceneParameter, SearchSceneParameter


@define(eq=False, slots=False)
class LambertianBSDF(BSDF):
    """
    Lambertian BSDF [``lambertian``].

    This class implements the Lambertian (a.k.a. diffuse) reflectance model.
    A surface with this scattering model attached scatters radiation equally in
    every direction.
    """

    reflectance: Spectrum = documented(
        attrs.field(
            default=0.5,
            converter=spectrum_factory.converter("reflectance"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("reflectance"),
            ],
        ),
        doc="Reflectance spectrum. Can be initialized with a dictionary "
        "processed by :data:`.spectrum_factory`.",
        type=".Spectrum",
        init_type=".Spectrum or dict or float",
        default="0.5",
    )

    @property
    def template(self) -> dict:
        # Inherit docstring
        result = {
            "type": "diffuse",
            **{
                f"reflectance.{key}": value
                for key, value in traverse(self.reflectance)[0].items()
            },
        }

        if self.id is not None:
            result["id"] = self.id

        return result

    @property
    def params(self) -> dict[str, SceneParameter]:
        # Inherit docstring
        params = traverse(self.reflectance)[1].data

        result = {}
        for key, param in params.items():
            result[f"reflectance.{key}"] = attrs.evolve(
                param,
                search=SearchSceneParameter(
                    node_type=mi.BSDF,
                    node_id=self.id,
                    parameter_relpath=f"reflectance.{key}",
                )
                if self.id is not None
                else None,
            )

        return result



from __future__ import annotations

import importlib
import typing as t
from abc import ABC, abstractmethod
from typing import Mapping, Sequence

import attrs
import mitsuba as mi
import numpy as np
import pint
import pinttr
from pinttr.util import ensure_units

from .._factory import Factory
from ..attrs import define, documented, frozen
from ..exceptions import TraversalError
from ..kernel import KernelDict, KernelSceneParameterMap, SceneParameter
from ..units import unit_context_config as ucc
from ..units import unit_registry as ureg

# ------------------------------------------------------------------------------
#                           Scene element interface
# ------------------------------------------------------------------------------


@define(eq=False, slots=False)
class SceneElement(ABC):
    """
    Abstract base class for all scene elements.

    Warnings
    --------
    All subclasses *must* have a hash, thus ``eq`` must be ``False`` (see
    `attrs docs on hashing <https://www.attrs.org/en/stable/hashing.html>`__
    for a complete explanation). This is required in order to make it possible
    to use caching decorators on instance methods.

    Notes
    -----
    The default implementation of ``__attrs_post_init__()`` executes the
    :meth:`update` method.
    """

    id: str | None = documented(
        attrs.field(
            default=None,
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc="Identifier of the current scene element.",
        type="str or None",
        init_type="str, optional",
    )

    def __attrs_post_init__(self):
        self.update()

    @property
    def params(self) -> dict[str, SceneParameter] | None:
        """
        Returns
        -------
        dict[str, :class:`.SceneParameter`] or None
            A dictionary mapping parameter paths, consisting of dot-separated
            strings, to a corresponding update protocol.

        See Also
        --------
        :class:`.SceneParameter`, :class:`.KernelSceneParameterMap`
        """
        return None

    @abstractmethod
    def traverse(self, callback) -> None:
        """
        Traverse this scene element and collect kernel dictionary template and
        parameter update map contributions.

        Parameters
        ----------
        callback : SceneTraversal
            Callback data structure storing the collected data.
        """
        pass

    def update(self) -> None:
        """
        Enforce internal state consistency. This method should be called when
        fields are modified. It is automatically called as a post-init step.
        """
        # The default implementation is a no-op
        pass


@define(eq=False, slots=False)
class NodeSceneElement(SceneElement, ABC):
    """
    Abstract base class for scene elements which expand as a single Mitsuba
    scene tree node which can be described as a scene dictionary.
    """

    @property
    @abstractmethod
    def template(self) -> dict:
        """
        Kernel dictionary template contents associated with this scene element.

        Returns
        -------
        dict
            A flat dictionary mapping dot-separated strings describing the path
            of an item in the nested scene dictionary to values. Values may be
            objects which can be directly used by the :func:`mitsuba.load_dict`
            function, or :class:`.DictParameter` instances which must be
            rendered.

        See Also
        --------
        :class:`.DictParameter`, :class:`.KernelDict`
        """
        pass

    @property
    def objects(self) -> dict[str, NodeSceneElement] | None:
        """
        Map of child objects associated with this scene element.

        Returns
        -------
        dict
            A dictionary mapping object names to a corresponding object to be
            inserted in the Eradiate scene graph.
        """
        return None

    def traverse(self, callback: SceneTraversal) -> None:
        # Inherit docstring
        callback.put_template(self.template)

        if self.params is not None:
            callback.put_params(self.params)

        if self.objects is not None:
            for name, obj in self.objects.items():
                callback.put_object(name, obj)


@define(eq=False, slots=False)
class InstanceSceneElement(SceneElement, ABC):
    """
    Abstract base class for scene elements which represent a node in the Mitsuba
    scene graph, but can only be expanded to a Mitsuba object.
    """

    @property
    @abstractmethod
    def instance(self) -> mi.Object:
        """
        Mitsuba object which is represented by this scene element.

        Returns
        -------
        mitsuba.Object
        """

        pass

    def traverse(self, callback):
        # Inherit docstring
        callback.put_instance(self.instance)

        if self.params is not None:
            callback.put_params(self.params)


@define(eq=False, slots=False)
class CompositeSceneElement(SceneElement, ABC):
    """
    Abstract based class for scene elements which expand to multiple Mitsuba
    scene tree nodes.
    """

    @property
    def template(self) -> dict:
        """
        Kernel dictionary template contents associated with this scene element.

        Returns
        -------
        dict
            A flat dictionary mapping dot-separated strings describing the path
            of an item in the nested scene dictionary to values. Values may be
            objects which can be directly used by the :func:`mitsuba.load_dict`
            function, or :class:`.DictParameter` instances which must be
            rendered.

        See Also
        --------
        :class:`.DictParameter`, :class:`.KernelDict`
        """
        return {}

    @property
    def objects(self) -> dict[str, NodeSceneElement] | None:
        """
        Map of child objects associated with this scene element.

        Returns
        -------
        dict
            A dictionary mapping object names to a corresponding object to be
            inserted in the Eradiate scene graph.
        """
        return None

    def traverse(self, callback):
        # Inherit docstring
        callback.put_template(self.template)

        if self.params is not None:
            callback.put_params(self.params)

        if self.objects is not None:
            for name, obj in self.objects.items():
                if isinstance(obj, InstanceSceneElement):
                    callback.put_object(name, obj)

                else:
                    template, params = traverse(obj)
                    callback.put_template(
                        {f"{name}.{k}": v for k, v in template.items()}
                    )
                    callback.put_params({f"{name}.{k}": v for k, v in params.items()})


@define(eq=False, slots=False)
class Ref(NodeSceneElement):
    """
    A scene element which represents a reference to a Mitsuba scene tree node.
    """

    id: str = documented(
        attrs.field(
            kw_only=True,
            validator=attrs.validators.instance_of(str),
        ),
        doc="Identifier of the referenced kernel scene object (required).",
        type="str",
    )

    @property
    def template(self) -> dict:
        # Inherit docstring
        return {"type": "ref", "id": self.id}


@define(eq=False, slots=False)
class Scene(NodeSceneElement):
    """
    A generic scene element container which expands as a :class:`mitsuba.Scene`
    object.
    """

    _objects: dict[str, SceneElement] = documented(
        attrs.field(factory=dict, converter=dict),
        doc="A map of scene elements which will be included in the Mitsuba "
        "scene definition.",
        type="dict",
        default="{}",
    )

    @property
    def template(self) -> dict:
        # Inherit docstring
        return {"type": "scene"}

    @property
    def objects(self) -> dict[str, SceneElement]:
        # Inherit docstring
        return self._objects


# ------------------------------------------------------------------------------
#                         Scene element tree traversal
# ------------------------------------------------------------------------------


@define
class SceneTraversal:
    """
    Data structure used to collect kernel dictionary data during scene element
    traversal.
    """

    #: Current traversal node
    node: NodeSceneElement

    #: Parent to current node
    parent: NodeSceneElement | None = attrs.field(default=None)

    #: Current node's name
    name: str | None = attrs.field(default=None)

    #: Current depth
    depth: int = attrs.field(default=0)

    #: Dictionary mapping nodes to their parents
    hierarchy: dict = attrs.field(factory=dict)

    #: Kernel dictionary template
    template: dict = attrs.field(factory=dict)

    #: Dictionary mapping nodes to their defined parameters
    params: dict = attrs.field(factory=dict)

    def __attrs_post_init__(self):
        self.hierarchy[self.node] = (self.parent, self.depth)

    def put_template(self, template: t.Mapping) -> None:
        """
        Add a contribution to the kernel dictionary template.
        """
        prefix = "" if self.name is None else f"{self.name}."

        for k, v in template.items():
            self.template[f"{prefix}{k}"] = v

    def put_params(self, params: t.Mapping) -> None:
        """
        Add a contribution to the parameter map.
        """
        prefix = "" if self.name is None else f"{self.name}."

        for k, v in params.items():
            self.params[f"{prefix}{k}"] = v

    def put_object(self, name: str, node: SceneElement) -> None:
        """
        Add a child object to the template and parameter map.
        """
        if isinstance(node, CompositeSceneElement):
            node.traverse(self)

        else:
            if node is None or node in self.hierarchy:
                return

            cb = type(self)(
                node=node,
                parent=self.node,
                name=name if self.name is None else f"{self.name}.{name}",
                depth=self.depth + 1,
                hierarchy=self.hierarchy,
                template=self.template,
                params=self.params,
            )

            if isinstance(node, InstanceSceneElement):
                cb.put_instance(node.instance)
                cb.put_params(node.params)

            else:
                node.traverse(cb)

    def put_instance(self, obj: mi.Object) -> None:
        """
        Add an instance to the kernel dictionary template.
        """
        if not self.name:
            raise TraversalError("Instances may only be inserted as child nodes.")
        self.template[self.name] = obj


def traverse(node: NodeSceneElement) -> tuple[KernelDict, KernelSceneParameterMap]:
    """
    Traverse a scene element tree and collect kernel dictionary template and
    parameter update table data.

    Parameters
    ----------
    node : .SceneElement
        Scene element where to start traversal.

    Returns
    -------
    kdict_template : .KernelDict
        Kernel dictionary template corresponding to the traversed scene element.

    umap_template : .KernelSceneParameterMap
        Kernel parameter table associated with the traversed scene element.
    """
    # Traverse scene element tree
    cb = SceneTraversal(node)
    node.traverse(cb)

    # Use collected data to generate the kernel dictionary
    return KernelDict(cb.template), KernelSceneParameterMap(cb.params)


# -- Misc (to be moved elsewhere) ----------------------------------------------


@frozen
class BoundingBox:
    """
    A basic data class representing an axis-aligned bounding box with
    unit-valued corners.

    Notes
    -----
    Instances are immutable.
    """

    min: pint.Quantity = documented(
        pinttr.field(
            units=ucc.get("length"),
            on_setattr=None,  # frozen instance: on_setattr must be disabled
        ),
        type="quantity",
        init_type="array-like or quantity",
        doc="Min corner.",
    )

    max: pint.Quantity = documented(
        pinttr.field(
            units=ucc.get("length"),
            on_setattr=None,  # frozen instance: on_setattr must be disabled
        ),
        type="quantity",
        init_type="array-like or quantity",
        doc="Max corner.",
    )

    @min.validator
    @max.validator
    def _min_max_validator(self, attribute, value):
        if not self.min.shape == self.max.shape:
            raise ValueError(
                f"while validating {attribute.name}: 'min' and 'max' must "
                f"have the same shape (got {self.min.shape} and {self.max.shape})"
            )
        if not np.all(np.less_equal(self.min, self.max)):
            raise ValueError(
                f"while validating {attribute.name}: 'min' must be less or "
                "equal to 'max'"
            )

    @classmethod
    def convert(
        cls, value: t.Sequence | t.Mapping | np.typing.ArrayLike | pint.Quantity
    ) -> t.Any:
        """
        Attempt conversion of a value to a :class:`BoundingBox`.

        Parameters
        ----------
        value
            Value to convert.

        Returns
        -------
        any
            If `value` is an array-like, a quantity or a mapping, conversion will
            be attempted. Otherwise, `value` is returned unmodified.
        """
        if isinstance(value, (np.ndarray, pint.Quantity)):
            return cls(value[0, :], value[1, :])

        elif isinstance(value, Sequence):
            return cls(*value)

        elif isinstance(value, Mapping):
            return cls(**pinttr.interpret_units(value, ureg=ureg))

        else:
            return value

    @property
    def shape(self):
        """
        tuple: Shape of `min` and `max` arrays.
        """
        return self.min.shape

    @property
    def extents(self) -> pint.Quantity:
        """
        :class:`pint.Quantity`: Extent in all dimensions.
        """
        return self.max - self.min

    @property
    def units(self):
        """
        :class:`pint.Unit`: Units of `min` and `max` arrays.
        """
        return self.min.units

    def contains(self, p: np.typing.ArrayLike, strict: bool = False) -> bool:
        """
        Test whether a point lies within the bounding box.

        Parameters
        ----------
        p : quantity or array-like
            An array of shape (3,) (resp. (N, 3)) representing one (resp. N)
            points. If a unitless value is passed, it is interpreted as
            ``ucc['length']``.

        strict : bool
            If ``True``, comparison is done using strict inequalities (<, >).

        Returns
        -------
        result : array of bool or bool
            ``True`` iff ``p`` in within the bounding box.
        """
        p = np.atleast_2d(ensure_units(p, ucc.get("length")))

        cmp = (
            np.logical_and(p > self.min, p < self.max)
            if strict
            else np.logical_and(p >= self.min, p <= self.max)
        )

        return np.all(cmp, axis=1)


# ------------------------------------------------------------------------------
#                               Factory accessor
# ------------------------------------------------------------------------------

_FACTORIES = {
    "atmosphere": "atmosphere.atmosphere_factory",
    "biosphere": "biosphere.biosphere_factory",
    "bsdf": "bsdfs.bsdf_factory",
    "illumination": "illumination.illumination_factory",
    "integrator": "integrators.integrator_factory",
    "measure": "measure.measure_factory",
    "phase": "phase.phase_function_factory",
    "shape": "shapes.shape_factory",
    "spectrum": "spectra.spectrum_factory",
    "surface": "surface.surface_factory",
}


def get_factory(element_type: str) -> Factory:
    """
    Return the factory corresponding to a scene element type.

    Parameters
    ----------
    element_type : str
        String identity of the scene element type associated to the requested
        factory.

    Returns
    -------
    factory : Factory
        Factory corresponding to the requested scene element type.

    Raises
    ------
    ValueError
        If the requested scene element type is unknown.

    Notes
    -----
    The ``element_type`` argument value maps to factories as follows:

    .. list-table::
       :widths: 1 1
       :header-rows: 1

       * - Element type ID
         - Factory
       * ``"atmosphere"``
         - :attr:`atmosphere_factory`
       * ``"biosphere"``
         - :attr:`biosphere_factory`
       * ``"bsdf"``
         - :attr:`bsdf_factory`
       * ``"illumination"``
         - :attr:`illumination_factory`
       * ``"integrator"``
         - :attr:`integrator_factory`
       * ``"measure"``
         - :attr:`measure_factory`
       * ``"phase"``
         - :attr:`phase_function_factory`
       * ``"shape"``
         - :attr:`shape_factory`
       * ``"spectrum"``
         - :attr:`spectrum_factory`
       * ``"surface"``
         - :attr:`surface_factory`
    """
    try:
        path = f"eradiate.scenes.{_FACTORIES[element_type]}"
    except KeyError:
        raise ValueError(
            f"unknown scene element type '{element_type}' "
            f"(should be one of {set(_FACTORIES.keys())})"
        )

    mod_path, attr = path.rsplit(".", 1)
    return getattr(importlib.import_module(mod_path), attr)



from __future__ import annotations

import logging
import typing as t

import attrs

from ._core import EarthObservationExperiment
from ._helpers import (
    check_geometry_atmosphere,
    check_path_compatible,
    check_piecewise_compatible,
    measure_inside_atmosphere,
    surface_converter,
)
from .. import validators
from ..attrs import AUTO, define, documented
from ..scenes.atmosphere import (
    Atmosphere,
    HeterogeneousAtmosphere,
    HomogeneousAtmosphere,
    MolecularAtmosphere,
    atmosphere_factory,
)
from ..scenes.biosphere import Canopy, biosphere_factory
from ..scenes.bsdfs import LambertianBSDF
from ..scenes.core import SceneElement
from ..scenes.geometry import (
    PlaneParallelGeometry,
    SceneGeometry,
)
from ..scenes.integrators import (
    PathIntegrator,
    PiecewiseVolPathIntegrator,
    VolPathIntegrator,
)
from ..scenes.measure import AbstractDistantMeasure, Measure
from ..scenes.shapes import RectangleShape
from ..scenes.surface import BasicSurface, CentralPatchSurface
from ..units import unit_context_config as ucc
from ..units import unit_registry as ureg

logger = logging.getLogger(__name__)


@define
class CanopyAtmosphereExperiment(EarthObservationExperiment):
    """
    Simulate radiation in a scene with an explicit canopy and atmosphere.
    This experiment assumes that the surface is plane and accounts for ground
    unit cell padding.

    Warnings
    --------
    * Canopy padding is controlled using the `padding` parameter: do *not* pad
      the canopy itself manually.

    Notes
    -----
    * A post-initialization step will constrain the measure setup if a
      distant measure is used and no target is defined:

      * if a canopy is defined, the target will be set to the top of the canopy
        unit cell (*i.e.* without its padding);
      * if no canopy is defined, the target will be set according to the
        atmosphere (*i.e.* to [0, 0, `toa`] where `toa` is the top-of-atmosphere
        altitude);
      * if neither atmosphere nor canopy are defined, the target is set to
        [0, 0, 0].

    * This experiment supports arbitrary measure positioning, except for
      :class:`.MultiRadiancemeterMeasure`, for which subsensor origins are
      required to be either all inside or all outside of the atmosphere. If an
      unsuitable configuration is detected, a :class:`ValueError` will be raised
      during initialization.

    * Currently this experiment is limited to the plane-parallel geometry.
    """

    # Currently, only the plane parallel geometry is supported
    geometry: PlaneParallelGeometry = documented(
        attrs.field(
            default="plane_parallel",
            converter=SceneGeometry.convert,
            validator=attrs.validators.instance_of(PlaneParallelGeometry),
        ),
        doc="Problem geometry.",
        type=".PlaneParallelGeometry",
        init_type="str or dict or .PlaneParallelGeometry",
        default='"plane_parallel"',
    )

    atmosphere: Atmosphere | None = documented(
        attrs.field(
            factory=HomogeneousAtmosphere,
            converter=attrs.converters.optional(atmosphere_factory.convert),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(Atmosphere)
            ),
        ),
        doc="Atmosphere specification. If set to ``None``, no atmosphere will "
        "be added. "
        "This parameter can be specified as a dictionary which will be "
        "interpreted by :data:`.atmosphere_factory`.",
        type=".Atmosphere or None",
        init_type=".Atmosphere or dict or None, optional",
        default=":class:`HomogeneousAtmosphere() <.HomogeneousAtmosphere>`",
    )

    canopy: Canopy | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(biosphere_factory.convert),
            validator=attrs.validators.optional(attrs.validators.instance_of(Canopy)),
        ),
        doc="Canopy specification. "
        "This parameter can be specified as a dictionary which will be "
        "interpreted by :data:`.biosphere_factory`.",
        type=".Canopy or None",
        init_type=".Canopy or dict or None, optional",
        default="None",
    )

    padding: int = documented(
        attrs.field(default=0, converter=int, validator=validators.is_positive),
        doc="Padding level. The canopy will be padded with copies to account for "
        "adjacency effects. This, in practice, has effects similar to "
        "making the scene periodic."
        "A value of 0 will yield only the defined scene. A value of 1 "
        "will add one copy in every direction, yielding a 3×3 patch. A "
        "value of 2 will yield a 5×5 patch, etc. The optimal padding level "
        "depends on the scene.",
        type="int",
        init_type="int, optional",
        default="0",
    )

    surface: BasicSurface | CentralPatchSurface | None = documented(
        attrs.field(
            factory=lambda: BasicSurface(bsdf=LambertianBSDF()),
            converter=attrs.converters.optional(surface_converter),
            validator=attrs.validators.optional(
                attrs.validators.instance_of((BasicSurface, CentralPatchSurface))
            ),
        ),
        doc="Surface specification. A :class:`.Surface` object may be passed: "
        "its shape specifications will be bypassed and the surface size will "
        "be computed automatically upon kernel dictionary generation. "
        "A :class:`.BSDF` may also be passed: it will be wrapped automatically "
        "in a :class:`.BasicSurface` instance. If a dictionary is passed, it "
        "will be first interpreted as a :class:`.BSDF`; if this fails, it will "
        "then be interpreted as a :class:`.Surface`. Finally, this field can "
        "be set to ``None``: in that case, no surface will be added.",
        type=".Surface or None",
        init_type=".Surface or .BSDF or dict or None, optional",
        default=":class:`BasicSurface(bsdf=LambertianBSDF()) <.BasicSurface>`",
    )

    def __attrs_post_init__(self):
        self._normalize_spectral()
        self._normalize_atmosphere()
        self._normalize_measures()
        self._normalize_integrator()

    def _normalize_atmosphere(self) -> None:
        """
        Enforce the experiment geometry on the atmosphere component(s).
        """
        if self.atmosphere is not None:
            # Since 'MolecularAtmosphere' cannot evaluate outside of its
            # vertical extent, we verify here that the experiment's geometry
            # comply with the atmosphere's vertical extent.
            if isinstance(self.atmosphere, MolecularAtmosphere):
                check_geometry_atmosphere(self.geometry, self.atmosphere)
            if isinstance(self.atmosphere, HeterogeneousAtmosphere):
                if self.atmosphere.molecular_atmosphere is not None:
                    check_geometry_atmosphere(
                        self.geometry, self.atmosphere.molecular_atmosphere
                    )

            # Override atmosphere geometry with experiment geometry
            self.atmosphere.geometry = self.geometry

            # The below call to update is required in the case of a
            # HeterogeneousAtmosphere, as it will propagate the geometry
            # override to its components.
            self.atmosphere.update()

    def _normalize_measures(self) -> None:
        """
        Ensure that distant measure targets are set to appropriate values.
        Processed measures will have their ray target and origin parameters
        overridden if relevant.
        """
        for measure in self.measures:
            # Override ray target location if relevant
            if isinstance(measure, AbstractDistantMeasure):
                if measure.target is None:
                    if self.canopy is None:  # No canopy: target origin point
                        measure.target = {"type": "point", "xyz": [0, 0, 0]}

                    else:  # Canopy: target top of canopy
                        measure.target = {
                            "type": "rectangle",
                            "xmin": -0.5 * self.canopy.size[0],
                            "xmax": 0.5 * self.canopy.size[0],
                            "ymin": -0.5 * self.canopy.size[1],
                            "ymax": 0.5 * self.canopy.size[1],
                            "z": self.canopy.size[2],
                        }

    def _normalize_integrator(self) -> None:
        """
        Ensures that the integrator is compatible with the atmosphere and geometry.
        """
        piecewise_compatible, pw_msg = check_piecewise_compatible(
            self.geometry, self.atmosphere
        )
        path_compatible, path_msg = check_path_compatible(self.atmosphere)

        if self.integrator is AUTO:
            if piecewise_compatible:
                self.integrator = PiecewiseVolPathIntegrator()
            else:
                self.integrator = VolPathIntegrator()
        else:
            msg = ""
            if (
                isinstance(self.integrator, PiecewiseVolPathIntegrator)
                and not piecewise_compatible
            ):
                msg = pw_msg

            if isinstance(self.integrator, PathIntegrator) and not path_compatible:
                msg = path_msg

            if msg:
                raise ValueError(msg)

    def _dataset_metadata(self, measure: Measure) -> dict[str, str]:
        result = super()._dataset_metadata(measure)

        if measure.is_distant():
            result["title"] = "Top-of-atmosphere simulation results"

        return result

    def _context_kwargs(self) -> dict[str, t.Any]:
        kwargs = {}

        for measure in self.measures:
            if measure_inside_atmosphere(self.atmosphere, measure):
                kwargs[f"{measure.sensor_id}.atmosphere_medium_id"] = (
                    self.atmosphere.medium_id
                )

        return kwargs

    @property
    def _default_surface_width(self):
        return 10.0 * ucc.get("length")

    @property
    def scene_objects(self) -> dict[str, SceneElement]:
        # Inherit docstring

        objects = {}

        # Note: Object size computation logic
        # - The atmosphere, if set, must be the largest object in the
        #   scene. If the geometry setup defines the atmosphere width, it is
        #   used. Otherwise, a size is computed automatically.
        # - The canopy size must be lower than the atmosphere size if it is
        #   defined.
        # - The surface must be larger than the largest object in the scene.
        #   If the atmosphere is set, the surface matches its size.
        #   Otherwise, if the canopy is set, the surface matches its size.
        #   Otherwise, the surface defaults to a size possibly specified by the
        #   geometry setup.

        # Pre-process atmosphere
        if self.atmosphere is not None:
            atmosphere = attrs.evolve(self.atmosphere, geometry=self.geometry)
            atmosphere_width = self.geometry.width
        else:
            atmosphere = None
            atmosphere_width = 0.0 * ureg.m

        # Pre-process canopy
        if self.canopy is not None:
            canopy_width = max(self.canopy.size[:2])

            if self.padding > 0:  # We must add extra instances if padding is requested
                canopy_width *= 2.0 * self.padding + 1.0
                canopy = self.canopy.padded_copy(self.padding)
            else:
                canopy = self.canopy
        else:
            canopy = None
            canopy_width = 0.0 * ureg.m

        # Check sizes, compute surface size
        if atmosphere is not None:
            assert atmosphere_width > canopy_width
        surface_width = self._default_surface_width
        if canopy_width > surface_width:
            surface_width = canopy_width
        if atmosphere_width > surface_width:
            surface_width = atmosphere_width

        # Pre-process surface
        if self.surface is not None:
            altitude = (
                atmosphere.bottom_altitude if atmosphere is not None else 0.0 * ureg.km
            )
            surface = attrs.evolve(
                self.surface,
                shape=RectangleShape.surface(altitude=altitude, width=surface_width),
            )
        else:
            surface = None

        # Add all configured elements to the scene
        if atmosphere is not None:
            objects["atmosphere"] = atmosphere

        if canopy is not None:
            objects["canopy"] = canopy

        if surface is not None:
            objects["surface"] = surface

        objects.update(
            {
                "illumination": self.illumination,
                **{measure.id: measure for measure in self.measures},
                "integrator": self.integrator,
            }
        )

        return objects



from __future__ import annotations

import attrs
import mitsuba as mi
import pint
import pinttrs

from ._core import BSDF
from ..core import traverse
from ..spectra import Spectrum, spectrum_factory
from ... import validators
from ...attrs import define, documented
from ...kernel import DictParameter, SceneParameter, SearchSceneParameter
from ...units import unit_registry as ureg


@define(eq=False, slots=False)
class OceanGraspBSDF(BSDF):
    """
    Ocean GRASP BSDF [``ocean_grasp``].

    This plugin implements the polarized ocean surface model as implemented
    by :cite:t:`Litvinov2024AerosolSurfaceCharacterization`. This model treats
    the ocean as an opaque surface and models the polarized sunglint, whitecap
    reflectance and underlight reflectance. It depends on wind speed and the
    index of refraction of the water and external medium (assumed to be air).

    See Also
    --------
    :ref:`plugin-bsdf-ocean_mishchenko`
    """

    wind_speed: pint.Quantity = documented(
        pinttrs.field(
            units=ureg("m/s").units,
            factory=lambda: 0.01 * ureg("m/s"),
            validator=[validators.is_positive, pinttrs.validators.has_compatible_units],
        ),
        doc="Wind speed [m/s] at 10 meters above the surface.",
        type="quantity",
        init_type="quantity or float",
        default="0.01 m/s",
    )

    eta: Spectrum = documented(
        attrs.field(
            default=1.33,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Real component of the water's index of refraction "
        "processed by :data:`.spectrum_factory`.",
        type=".Spectrum",
        init_type=".Spectrum or dict or float",
        default="1.33",
    )

    k: Spectrum = documented(
        attrs.field(
            default=0.0,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Imaginary component of the water's index of refraction "
        "processed by :data:`.spectrum_factory`.",
        type=".Spectrum",
        init_type=".Spectrum or dict or float",
        default="0.0",
    )

    ext_ior: Spectrum = documented(
        attrs.field(
            default=1.0,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Real component of the air's index of refraction "
        "processed by :data:`.spectrum_factory`. The imaginary component "
        "is assumed to be zero. ",
        type=".Spectrum",
        init_type=".Spectrum or dict or float",
        default="1.000277",
    )

    water_body_reflectance: Spectrum = documented(
        attrs.field(
            default=0.0,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Diffuse reflectance of radiations that underwent one or more  "
        "scattering events within the water body.",
        type=".Spectrum",
        init_type=".Spectrum or dict or float",
        default="0.0",
    )

    @property
    def template(self) -> dict:
        # Inherit docstring
        objects = {
            "eta": traverse(self.eta)[0],
            "k": traverse(self.k)[0],
            "ext_ior": traverse(self.ext_ior)[0],
            "water_body_reflectance": traverse(self.water_body_reflectance)[0],
        }

        result = {
            "type": "ocean_grasp",
            "wavelength": DictParameter(lambda ctx: ctx.si.w.m_as("nm")),
            "wind_speed": self.wind_speed.m_as("m/s"),
        }

        for obj_key, obj_values in objects.items():
            for key, value in obj_values.items():
                result[f"{obj_key}.{key}"] = value

        if self.id is not None:
            result["id"] = self.id

        return result

    @property
    def params(self) -> dict[str, SceneParameter]:
        # Inherit docstring
        objects = {
            "eta": traverse(self.eta)[1],
            "k": traverse(self.k)[1],
            "ext_ior": traverse(self.ext_ior)[1],
            "water_body_reflectance": traverse(self.water_body_reflectance)[1],
        }

        result = {}
        for obj_key, obj_params in objects.items():
            for key, param in obj_params.items():
                result[f"{obj_key}.{key}"] = attrs.evolve(
                    param,
                    search=SearchSceneParameter(
                        node_type=mi.BSDF,
                        node_id=self.id,
                        parameter_relpath=f"{obj_key}.{key}",
                    )
                    if self.id is not None
                    else None,
                )

        result["wavelength"] = SceneParameter(lambda ctx: ctx.si.w.m_as("nm"))
        return result



from importlib.metadata import PackageNotFoundError, version

try:
    _version = version("eradiate")
except PackageNotFoundError as e:
    raise PackageNotFoundError(
        "Eradiate is not installed; please install it in your Python environment."
    ) from e



"""
Upon import, this module checks if the kernel dependencies are correctly set up.
"""

from importlib.metadata import PackageNotFoundError, version

# Internal constants
REQUIRED_MITSUBA_VERSION = "3.5.2"
REQUIRED_MITSUBA_PATCH_VERSION = "0.3.2"


def find_drjit():
    # Check if Dr.Jit and Mitsuba can be imported successfully
    try:
        __import__("drjit")
        return True

    except (ImportError, ModuleNotFoundError) as e:
        raise ImportError(
            "Could not import module 'drjit'. If you are an end-user, you need "
            "to install the radiometric kernel by adding the 'kernel' extra to "
            "the installation command (e.g. pip install 'eradiate[kernel]'). "
            "If you are a developer, you have to build the kernel and add "
            "it to your $PYTHONPATH."
        ) from e


def find_mitsuba():
    # Check if Dr.Jit and Mitsuba can be imported successfully
    try:
        __import__("mitsuba")
        return True

    except (ImportError, ModuleNotFoundError) as e:
        raise ImportError(
            "Could not import module 'mitsuba'. If you are an end-user, you need "
            "to install the radiometric kernel by adding the 'kernel' extra to "
            "the installation command (e.g. pip install 'eradiate[kernel]'). "
            "If you are a developer, you have to build the kernel and add "
            "it to your $PYTHONPATH."
        ) from e


def kernel_version():
    """
    Detect the version of the Eradiate kernel.

    Returns
    -------
    mitsuba_version : str
        The version of Mitsuba on which the kernel is based. This is the
        :data:`mitsuba.MI_VERSION` constant.

    patch_version : str or None
        The version of the kernel. This is the :data:`mitsuba.ERD_MI_VERSION`
        constant. If the constant is not found, *i.e.* if the detected Mitsuba
        package is not the Eradiate kernel, ``None`` is returned.

    Notes
    -----
    This function reads the metadata exposed by the ``mitsuba`` module, not
    package metadata.
    """
    # Retrieve the kernel and kernel patch versions
    _mi = __import__("mitsuba")
    mitsuba_version = _mi.scalar_rgb.MI_VERSION
    patch_version = (
        _mi.scalar_rgb.ERD_MI_VERSION
        if hasattr(_mi.scalar_rgb, "ERD_MI_VERSION")
        else None
    )
    return mitsuba_version, patch_version


def kernel_installed():
    """
    Check if the kernel Python package is installed to the active Python
    environment.

    Returns
    -------
    is_installed : bool
        ``True`` iff ``eradiate-mitsuba`` package is installed to the active
        Python environment.

    package_version : str or None
         If ``is_installed`` is ``True``, a string containing the version found
         for that package.
    """

    # Check if the eradiate-mitsuba package is installed
    try:
        return True, version("eradiate-mitsuba")
    except PackageNotFoundError:
        return False, None


def check_kernel():
    """
    Perform a few basic checks on the kernel.
    """
    warnings = []

    # Can we find Dr.Jit in the current Python path?
    find_drjit()

    # Can we find Mitsuba in the current Python path?
    find_mitsuba()

    # Is the kernel version compatible with requirements?
    mitsuba_version, patch_version = kernel_version()

    if patch_version is None:
        warnings.append(
            "Detected Mitsuba build is not the Eradiate Mitsuba version. "
            "Make sure to use the Eradiate-specific version of Mitsuba, compiled "
            "with the appropriate variants and plugins."
        )
    elif (
        patch_version != REQUIRED_MITSUBA_PATCH_VERSION
        or mitsuba_version != REQUIRED_MITSUBA_VERSION
    ):
        warnings.append(
            f"Found incompatible kernel version "
            f"{patch_version}, based on Mitsuba {mitsuba_version}. "
            f"Required: {REQUIRED_MITSUBA_PATCH_VERSION}, based on Mitsuba "
            f"{REQUIRED_MITSUBA_VERSION}."
        )

    return warnings


check_kernel()



from __future__ import annotations

import attrs
import mitsuba as mi

from ._core import BSDF
from ..core import traverse
from ..spectra import Spectrum, spectrum_factory
from ... import validators
from ...attrs import define, documented
from ...kernel import SceneParameter, SearchSceneParameter


@define(eq=False, slots=False)
class RPVBSDF(BSDF):
    """
    RPV BSDF [``rpv``].

    This BSDF implements the Rahman-Pinty-Verstraete (RPV) reflection model
    :cite:`Rahman1993CoupledSurfaceatmosphereReflectance,Pinty2000SurfaceAlbedoRetrieval`.
    It notably features a controllable back-scattering lobe (`hot spot`)
    characteristic of many natural land surfaces and is frequently used in Earth
    observation because of its simple parametrisation.

    See Also
    --------
    :ref:`plugin-bsdf-rpv`

    Notes
    -----
    * The default configuration is typical of grassland in the visible domain
      (:cite:`Rahman1993CoupledSurfaceatmosphereReflectance`, Table 1).
    * Parameter names are defined as per the symbols used in the Eradiate
      Scientific Handbook :cite:`EradiateScientificHandbook2020`.
    """

    rho_0: Spectrum = documented(
        attrs.field(
            default=0.183,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Amplitude parameter. Must be dimensionless. Should be in :math:`[0, 1]`.",
        type=".Spectrum",
        init_type=".Spectrum or dict or float, optional",
        default="0.183",
    )

    rho_c: Spectrum | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(
                spectrum_factory.converter("dimensionless")
            ),
            validator=attrs.validators.optional(
                [
                    attrs.validators.instance_of(Spectrum),
                    validators.has_quantity("dimensionless"),
                ]
            ),
        ),
        doc="Hot spot parameter. Must be dimensionless. "
        r"Should be in :math:`[0, 1]`. If unset, :math:`\rho_\mathrm{c}` "
        r"defaults to the kernel plugin default (equal to :math:`\rho_0`).",
        type=".Spectrum or None",
        init_type=".Spectrum or dict or float or None, optional",
        default="None",
    )

    k: Spectrum = documented(
        attrs.field(
            default=0.780,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Bowl-shape parameter. Must be dimensionless. Should be in :math:`[0, 2]`.",
        type=".Spectrum",
        init_type=".Spectrum or dict or float, optional",
        default="0.780",
    )

    g: Spectrum = documented(
        attrs.field(
            default=-0.1,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Asymmetry parameter. Must be dimensionless. Should be in :math:`[-1, 1]`.",
        type=".Spectrum",
        init_type=".Spectrum or dict or float, optional",
        default="-0.1",
    )

    @property
    def template(self) -> dict:
        # Inherit docstring
        objects = {
            "rho_0": traverse(self.rho_0)[0],
            "k": traverse(self.k)[0],
            "g": traverse(self.g)[0],
        }

        if self.rho_c is not None:
            objects["rho_c"] = traverse(self.rho_c)[0]

        result = {"type": "rpv"}

        for obj_key, obj_values in objects.items():
            for key, value in obj_values.items():
                result[f"{obj_key}.{key}"] = value

        if self.id is not None:
            result["id"] = self.id

        return result

    @property
    def params(self) -> dict[str, SceneParameter]:
        # Inherit docstring
        objects = {
            "rho_0": traverse(self.rho_0)[1],
            "k": traverse(self.k)[1],
            "g": traverse(self.g)[1],
        }

        if self.rho_c is not None:
            objects["rho_c"] = traverse(self.rho_c)[1]

        result = {}
        for obj_key, obj_params in objects.items():
            for key, param in obj_params.items():
                result[f"{obj_key}.{key}"] = attrs.evolve(
                    param,
                    search=SearchSceneParameter(
                        node_type=mi.BSDF,
                        node_id=self.id,
                        parameter_relpath=f"{obj_key}.{key}",
                    )
                    if self.id is not None
                    else None,
                )

        return result



from . import config as config
from . import constants as constants
from . import contexts as contexts
from . import converters as converters
from . import data as data
from . import experiments as experiments
from . import frame as frame
from . import kernel as kernel
from . import notebook as notebook
from . import pipelines as pipelines
from . import plot as plot
from . import radprops as radprops
from . import rng as rng
from . import scenes as scenes
from . import spectral as spectral
from . import units as units
from . import validators as validators
from . import xarray as xarray
from ._mode import Mode as Mode
from ._mode import ModeFlag as ModeFlag
from ._mode import get_mode as get_mode
from ._mode import mode as mode
from ._mode import modes as modes
from ._mode import set_mode as set_mode
from ._mode import supported_mode as supported_mode
from ._mode import unsupported_mode as unsupported_mode
from .contexts import KernelContext as KernelContext
from .data._asset_manager import asset_manager as asset_manager
from .data._file_resolver import fresolver as fresolver
from .experiments import run as run
from .notebook import load_ipython_extension as load_ipython_extension
from .scenes.core import traverse as traverse
from .units import unit_context_config as unit_context_config
from .units import unit_context_kernel as unit_context_kernel
from .units import unit_registry as unit_registry



from __future__ import annotations

from abc import ABC

from ..core import NodeSceneElement
from ..._factory import Factory
from ...attrs import define

bsdf_factory = Factory()
bsdf_factory.register_lazy_batch(
    [
        ("_black.BlackBSDF", "black", {}),
        ("_checkerboard.CheckerboardBSDF", "checkerboard", {}),
        ("_lambertian.LambertianBSDF", "lambertian", {}),
        ("_mqdiffuse.MQDiffuseBSDF", "mqdiffuse", {}),
        ("_ocean_legacy.OceanLegacyBSDF", "ocean_legacy", {}),
        ("_ocean_grasp.OceanGraspBSDF", "ocean_grasp", {}),
        ("_ocean_mishchenko.OceanMishchenkoBSDF", "ocean_mishchenko", {}),
        ("_opacity_mask.OpacityMaskBSDF", "opacity_mask", {}),
        ("_rpv.RPVBSDF", "rpv", {}),
        ("_rtls.RTLSBSDF", "rtls", {}),
        ("_hapke.HapkeBSDF", "hapke", {}),
    ],
    cls_prefix="eradiate.scenes.bsdfs",
)


@define(eq=False, slots=False)
class BSDF(NodeSceneElement, ABC):
    """
    Abstract base class  for all BSDF scene elements.
    """

    pass



from __future__ import annotations

import typing as t
from collections import abc as cabc

import attrs
import mitsuba as mi
import numpy as np

from ._core import PhaseFunction, phase_function_factory
from ..core import traverse
from ..geometry import PlaneParallelGeometry, SceneGeometry, SphericalShellGeometry
from ...attrs import documented
from ...contexts import KernelContext
from ...kernel import DictParameter, KernelSceneParameterFlags, SceneParameter
from ...spectral.index import SpectralIndex
from ...util.misc import cache_by_id


@attrs.define(eq=False, slots=False)
class BlendPhaseFunction(PhaseFunction):
    """
    Blended phase function [``blend_phase``].

    This phase function aggregates two or more sub-phase functions
    (*components*) and blends them based on its `weights` parameter. Weights are
    usually based on the associated medium's scattering coefficient.
    """

    components: list[PhaseFunction] = documented(
        attrs.field(
            converter=lambda x: [phase_function_factory.convert(y) for y in x],
            validator=attrs.validators.deep_iterable(
                attrs.validators.instance_of(PhaseFunction)
            ),
            kw_only=True,
        ),
        type="list of :class:`.PhaseFunction`",
        init_type="list of :class:`.PhaseFunction` or list of dict",
        doc="List of components (at least two). This parameter has not default.",
    )

    @components.validator
    def _components_validator(self, attribute, value):
        if not len(value) > 1:
            raise ValueError(
                f"while validating {attribute.name}: BlendPhaseFunction must "
                "have at least two components"
            )

    weights: np.ndarray | list[t.Callable[[KernelContext], np.ndarray]] = documented(
        attrs.field(
            converter=lambda x: x if callable(x[0]) else np.array(x, dtype=np.float64),
            kw_only=True,
        ),
        type="ndarray or list of callables",
        init_type="array-like or list of callables",
        doc="List of weights associated with each component. Weights may be "
        "numerical values; in that case, they must be of shape (n,) or (n, m), "
        "where n is the number of components and m the number of cells along "
        "the atmosphere's vertical axis. Alternatively, weights may be "
        "callables that take a :class:`.KernelContext` as argument and "
        "return an array of shape (n, m). "
        "This parameter is required and has no default.",
    )

    @weights.validator
    def _weights_validator(self, attribute, value):
        if isinstance(value, np.ndarray):
            if value.ndim == 0 or value.ndim > 2:
                raise ValueError(
                    f"while validating '{attribute.name}': array must have 1 or 2 "
                    f"dimensions, got {value.ndim}"
                )

            if not value.shape[0] == len(self.components):
                raise ValueError(
                    f"while validating '{attribute.name}': array must have shape "
                    "(n,) or (n, m) where n is the number of components; got "
                    f"{value.shape}"
                )

        elif isinstance(value, cabc.Sequence):
            if not len(value) == len(self.components):
                raise ValueError(
                    f"while validating '{attribute.name}': weight and component "
                    "lists must have the same length"
                )

    geometry: SceneGeometry | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(SceneGeometry.convert),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(SceneGeometry)
            ),
        ),
        doc="Parameters defining the basic geometry of the scene. If unset, "
        "the volume textures defining component weights will be assigned "
        "defaults likely unsuitable for atmosphere construction.",
        type=".SceneGeometry or None",
        init_type=".SceneGeometry or dict or str, optional",
        default="None",
    )

    def update(self) -> None:
        super().update()

        # Synchronize geometries
        for component in self.components:
            component.update()

            if isinstance(component, BlendPhaseFunction):
                component.geometry = self.geometry

    @cache_by_id
    def _eval_conditional_weights_impl(self, si: SpectralIndex) -> np.ndarray:
        """
        Memoised weight evaluation, used if weights are defined as callables.
        """
        n_comp = len(self.components)

        if isinstance(self.weights, list):
            weights = np.array([w(si) for w in self.weights], dtype=np.float64)
        else:  # if isinstance(self.weights, np.ndarray):
            weights = np.array(self.weights, dtype=np.float64)

        if weights.ndim < 2:
            weights = weights.reshape((-1, 1))

        result = np.zeros((n_comp - 1, *weights.shape[1:]), dtype=np.float64)

        # Compute conditional weights
        for i in range(n_comp - 1):
            # Normalize weights
            weights_sum = weights[i:, ...].sum(axis=0, keepdims=True)
            weights_normalized = np.divide(
                weights[i:, ...],
                weights_sum,
                where=weights_sum != 0.0,
                out=np.zeros_like(weights[i:, ...]),
            )
            # Aggregate weights of all components except the first one
            result[i] = weights_normalized[1:, ...].sum(axis=0, keepdims=True)

        return result

    def eval_conditional_weights(
        self,
        si: SpectralIndex,
        n_component: int | list[int] | None = None,
    ) -> np.ndarray:
        """
        Evaluate the conditional weights of specified Mitsuba phase function
        components.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral context.

        n_component : int or list of int, optional
            The index of the Mitsuba phase function component for which the
            conditional weight should be evaluated. If ``None``, the conditional
            weights of all components will be evaluated.

        Returns
        -------
        ndarray
            Conditional weights of the specified components as an array of shape
            (N, M) where n is the number of components and m the number of cells
            along the atmosphere's vertical axis.
        """
        if n_component is None:
            n_component = range(len(self.components) - 1)
        elif isinstance(n_component, int):
            n_component = [n_component]

        # Compute normalized component weights (cached until call with different
        # context)
        weights = self._eval_conditional_weights_impl(si)

        # Return selected components
        return weights[n_component, ...]

    @property
    def template(self) -> dict:
        result = {"type": "blendphase"}

        for i in range(len(self.components) - 1):
            prefix = "phase_1." * i

            # Add components
            template, _ = traverse(self.components[i])
            result.update(
                {
                    **{f"{prefix}phase_0.{k}": v for k, v in template.items()},
                    f"{prefix}phase_1.type": "blendphase",
                }
            )

            # Assign conditional weight to second component
            if self.geometry is None or isinstance(
                self.geometry, PlaneParallelGeometry
            ):
                # Note: This defines a partial and evaluates the component index.
                # Passing i as the kwarg default value is essential to force the
                # dereferencing of the loop variable.
                def eval_conditional_weights(ctx: KernelContext, n_component=i):
                    return mi.VolumeGrid(
                        np.reshape(
                            self.eval_conditional_weights(ctx.si, n_component),
                            (-1, 1, 1),  # Mind dim ordering! (C-style, i.e. zyx)
                        ).astype(np.float32)
                    )

                result[f"{prefix}weight.type"] = "gridvolume"
                result[f"{prefix}weight.grid"] = DictParameter(eval_conditional_weights)
                result[f"{prefix}weight.filter_type"] = "nearest"

                if self.geometry is not None:
                    result[f"{prefix}weight.to_world"] = (
                        self.geometry.atmosphere_volume_to_world
                    )

            elif isinstance(self.geometry, SphericalShellGeometry):
                # Same comment as above
                def eval_conditional_weights(ctx: KernelContext, n_component=i):
                    return mi.VolumeGrid(
                        np.reshape(
                            self.eval_conditional_weights(ctx.si, n_component),
                            (1, 1, -1),  # Mind dim ordering! (C-style, i.e. zyx)
                        ).astype(np.float32)
                    )

                result[f"{prefix}weight.type"] = "sphericalcoordsvolume"
                result[f"{prefix}weight.volume.type"] = "gridvolume"
                result[f"{prefix}weight.volume.grid"] = DictParameter(
                    eval_conditional_weights
                )
                result[f"{prefix}weight.volume.filter_type"] = "nearest"
                result[f"{prefix}weight.to_world"] = (
                    self.geometry.atmosphere_volume_to_world
                )
                result[f"{prefix}weight.rmin"] = self.geometry.atmosphere_volume_rmin

            else:
                raise ValueError(
                    f"unhandled scene geometry type '{type(self.geometry).__name__}'"
                )

        else:
            template, _ = traverse(self.components[-1])
            result.update({**{f"{prefix}phase_1.{k}": v for k, v in template.items()}})

        return result

    @property
    def params(self) -> dict[str, SceneParameter]:
        result = {}

        for i in range(len(self.components) - 1):
            prefix = "phase_1." * i

            # Add components
            _, params = traverse(self.components[i])
            result.update(
                {
                    **{f"{prefix}phase_0.{k}": v for k, v in params.items()},
                }
            )

            if self.geometry is None or isinstance(
                self.geometry, PlaneParallelGeometry
            ):
                # Note: This defines a partial and evaluates the component index.
                # Passing i as the kwarg default value is essential to force the
                # dereferencing of the loop variable.
                def eval_conditional_weights(ctx: KernelContext, n_component=i):
                    return np.reshape(
                        self.eval_conditional_weights(ctx.si, n_component),
                        (-1, 1, 1, 1),  # Mind dim ordering! (C-style, i.e. zyxc)
                    ).astype(np.float32)

                # Assign conditional weight to second component
                result[f"{prefix}weight.data"] = SceneParameter(
                    eval_conditional_weights,
                    KernelSceneParameterFlags.SPECTRAL,
                )

            elif isinstance(self.geometry, SphericalShellGeometry):
                # Same comment as above
                def eval_conditional_weights(ctx: KernelContext, n_component=i):
                    return np.reshape(
                        self.eval_conditional_weights(ctx.si, n_component),
                        (1, 1, -1, 1),  # Mind dim ordering! (C-style, i.e. zyxc)
                    ).astype(np.float32)

                # Assign conditional weight to second component
                result[f"{prefix}weight.volume.data"] = SceneParameter(
                    eval_conditional_weights,
                    KernelSceneParameterFlags.SPECTRAL,
                )

            else:
                raise NotImplementedError

        else:
            _, params = traverse(self.components[-1])
            result.update({**{f"{prefix}phase_1.{k}": v for k, v in params.items()}})

        return result
